"""Class utilities to access to the DOM."""

import re
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Literal, overload

import bs4

if TYPE_CHECKING:
    from .testclient import WebResponse, WebTestClient  # coverage: ignore


class Element:
    """Access to a dom element."""

    def __init__(self, client: "WebTestClient", tag: bs4.Tag):
        self._client = client
        self._tag = tag

    def click(self) -> "WebResponse":
        """Simulate a client to a a link. No javascript exectuted here."""
        return self._client.get(self._tag.attrs["href"])

    @property
    def node_name(self) -> str:
        """Get the node name of the dom element."""
        return self._tag.name

    @property
    def attrs(self) -> dict[str, str]:
        """Attributes of the element."""
        return self._tag.attrs

    @property
    def text(self) -> str:
        """
        Return the text of the element, with text of childs element.

        Note that the text is stripped for convenience but inner text may contains
        many spaces not manipulated here.
        """
        return self._tag.text.strip()

    @property
    def h1(self) -> "Element":
        """
        Return the h1 child element.

        Should be used on the html body element directly.
        """
        nodes = self.by_node_name("h1")
        assert len(nodes) == 1, f"Should have 1 <h1>, got {len(nodes)} in {self}"
        return nodes[0]

    @property
    def h2(self) -> Sequence["Element"]:
        """
        Return the h2 elements.
        """
        return self.by_node_name("h2")

    @property
    def form(self) -> "Element | None":
        """Get the form element of the web page."""
        return Element(self._client, self._tag.form) if self._tag.form else None

    @property
    def hx_target(self) -> str | None:
        """
        Return the hx-target of the element.

        It may be set on a parent. It also resolve special case "this" and return the id
        of the element.
        """
        el: bs4.Tag | None = self._tag
        while el:
            if "hx-target" in el.attrs:
                ret = el.attrs["hx-target"]
                if ret == "this":
                    ret = el.attrs["id"]
                return ret
            el = el.parent
        return None

    def by_text(
        self, text: str, *, node_name: str | None = None, position: int | None = None
    ) -> "Element | None":
        """Find the first element that match the text."""
        nodes = self.iter_all_by_text(text, node_name=node_name)
        ret = list(nodes)
        if not ret:
            return None
        if position is None:
            assert len(ret) == 1, f"Should have 1 element, got {len(ret)} in {self}"
        else:
            assert len(ret) > position, "Not enough element found"
        return ret[position or 0]

    def iter_all_by_text(
        self, text: str, *, node_name: str | None = None
    ) -> "Iterator[Element]":
        """Return an iterator of all elements that match the text."""
        nodes = self._tag.find_all(string=re.compile(rf"\s*{text}\s*"))
        for node in nodes:
            if isinstance(node, bs4.NavigableString):
                node = node.parent

            if node_name:
                while node is not None:
                    if node.name == node_name:
                        yield Element(self._client, node)
                    node = node.parent
            elif node:
                yield Element(self._client, node)
        return None

    def get_all_by_text(
        self, text: str, *, node_name: str | None = None
    ) -> "Sequence[Element]":
        """Return the list of all elements that match the text."""
        nodes = self.iter_all_by_text(text, node_name=node_name)
        return list(nodes)

    def by_label_text(self, text: str) -> "Element | None":
        """Return the element which is the target of the label having the given text."""
        label = self.by_text(text, node_name="label")
        assert label is not None
        assert label.attrs.get("for") is not None
        resp = self._tag.find(id=label.attrs["for"])
        assert not isinstance(resp, bs4.NavigableString)
        return Element(self._client, resp) if resp else None

    def by_id(self, id: str) -> "Element | None":
        """Find the element having the given id."""
        resp = self._tag.find_all(id=id)
        assert not isinstance(resp, bs4.NavigableString)
        if not resp:
            return None
        assert len(resp) == 1
        return Element(self._client, resp[0]) if resp else None

    @overload
    def by_node_name(
        self,
        node_name: str,
        *,
        attrs: dict[str, str] | None = None,
        multiple: Literal[False],
    ) -> "Element": ...

    @overload
    def by_node_name(
        self,
        node_name: str,
        *,
        attrs: dict[str, str] | None = None,
        multiple: Literal[True],
    ) -> "list[Element]": ...

    @overload
    def by_node_name(
        self,
        node_name: str,
        *,
        attrs: dict[str, str] | None = None,
    ) -> "list[Element]": ...

    def by_node_name(
        self,
        node_name: str,
        *,
        attrs: dict[str, str] | None = None,
        multiple: bool = True,
    ) -> "list[Element] | Element":
        """
        Return the list of elements with the given node_name.

        An optional set of attributes may given and must match if passed.
        """
        ret = [
            Element(self._client, e) for e in self._tag.find_all(node_name, attrs or {})
        ]
        if not multiple:
            assert len(ret) == 1
            return ret[0]
        return ret

    def __repr__(self) -> str:
        return f"<{self.node_name}>"

    def __str__(self) -> str:
        return str(self._tag)



"""
Create markup for pydantic forms.
"""

import secrets
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, get_origin

from markupsafe import Markup
from pydantic.fields import FieldInfo

from fastlife.adapters.xcomponent.pydantic_form.widgets.base import CustomWidget, Widget
from fastlife.domain.model.form import FormModel
from fastlife.domain.model.template import XTemplate

if TYPE_CHECKING:
    from fastlife.service.templates import AbstractTemplateRenderer

from .base import BaseWidgetBuilder
from .bool_builder import BoolBuilder
from .emailstr_builder import EmailStrBuilder
from .enum_builder import EnumBuilder
from .literal_builder import LiteralBuilder
from .model_builder import ModelBuilder
from .secretstr_builder import SecretStrBuilder
from .sequence_builder import SequenceBuilder
from .set_builder import SetBuilder
from .simpletype_builder import SimpleTypeBuilder
from .union_builder import UnionBuilder


class OptionalFatalError(XTemplate):
    template = """
    <>
      {
        if message {
            <FatalError message={message} />
        }
      }
    </>
    """
    message: str


class WidgetFactory:
    """
    Form builder for pydantic model.

    :param renderer: template engine to render widget.
    :param token: reuse a token.
    """

    def __init__(self, renderer: "AbstractTemplateRenderer", token: str | None = None):
        self.renderer = renderer
        self.token = token or secrets.token_urlsafe(4).replace("_", "-")
        self.builders: list[BaseWidgetBuilder[Any]] = [
            # Order is super important here
            # starts by the union type
            UnionBuilder(self),
            # and to other types that have an origin
            SetBuilder(self),
            LiteralBuilder(self),
            SequenceBuilder(self),
            # from this part, order does not really matter
            ModelBuilder(self),
            BoolBuilder(self),
            EnumBuilder(self),
            EmailStrBuilder(self),
            SecretStrBuilder(self),
            # we keep simple types, str, int at the end
            SimpleTypeBuilder(self),
        ]

    def get_markup(
        self,
        model: FormModel[Any],
        *,
        removable: bool = False,
        field: FieldInfo | None = None,
    ) -> Markup:
        """
        Get the markup for the given model.

        :param model: the model to build the html markup.
        :param removable: Include a button to remove the model in the markup.
        :param field: only build the markup of this field is not None.
        """
        ret = Markup()
        if model.fatal_error:
            ret += Markup(
                self.renderer.render_template(
                    OptionalFatalError(message=model.fatal_error)
                )
            )
        ret += Markup(
            self.get_widget(
                model.model.__class__,
                model.form_data,
                model.errors,
                prefix=model.prefix,
                removable=removable,
                field=field,
            ).to_html(self.renderer)
        )
        return ret

    def get_widget(
        self,
        base: type[Any],
        form_data: Mapping[str, Any],
        form_errors: Mapping[str, Any],
        *,
        prefix: str,
        removable: bool,
        field: FieldInfo | None = None,
    ) -> Widget[Any]:
        """
        build the widget for the given type and return it.
        :param base: the type to build, it has to be a builtin or a Pydantic model.
        :param form_data: form values to render.
        :param form_errors: form errors to render.
        """
        return self.build(
            base,
            value=form_data.get(prefix, {}),
            form_errors=form_errors,
            name=prefix,
            removable=removable,
            field=field,
        )

    def build(
        self,
        typ: type[Any],
        *,
        name: str = "",
        value: Any,
        removable: bool,
        form_errors: Mapping[str, Any],
        field: FieldInfo | None = None,
    ) -> Widget[Any]:
        """
        build widget tree for the given type.
        This function is recurive and shoud not be used directly.
        The type is a composite, it can be pydantic model, builtin, list or unions.

        The {meth}`WidgetFactory.get_widget` or {meth}`WidgetFactory.get_markup`
        should be used.

        :param typ: the type to build, it has to be a builtin or a Pydantic model.
        :param name: name of the widget to build.
        :param value: value for the widget.
        :param removable: True if it has to include a remove button.
        :param form_errors: errors in the form.
        :param field: field information used to customize the widget.
        """
        if field and field.metadata:
            for widget in field.metadata:
                if isinstance(widget, CustomWidget):
                    ret: Widget[Any] = widget.typ(
                        name=name,
                        value=value,
                        removable=removable,
                        title=field.title or "" if field else "",
                        hint=field.description if field else None,
                        aria_label=(
                            field.json_schema_extra.get("aria_label")  # type:ignore
                            if field and field.json_schema_extra
                            else None
                        ),
                        token=self.token,
                        error=form_errors.get(name),
                    )
                    return ret

        type_origin = get_origin(typ)
        for builder in self.builders:
            if builder.accept(typ, type_origin):
                return builder.build(
                    field_name=name,
                    field_type=typ,
                    field=field,
                    value=value,
                    form_errors=form_errors,
                    removable=removable,
                )

        raise NotImplementedError(f"{typ} not implemented")  # coverage: ignore



"""Widget base class."""

import secrets
from collections.abc import Mapping
from typing import Any, Generic, Self, TypeVar

from markupsafe import Markup
from pydantic import Field, model_validator

from fastlife.domain.model.template import XTemplate
from fastlife.service.templates import AbstractTemplateRenderer
from fastlife.shared_utils.infer import is_union

T = TypeVar("T")


def get_title(typ: type[Any]) -> str:
    return getattr(
        getattr(typ, "__meta__", None),
        "title",
        getattr(typ, "__name__", ""),
    )


class Widget(XTemplate, Generic[T]):
    """
    Base class for widget of pydantic fields.

    :param name: field name.
    :param value: field value.
    :param title: title for the widget.
    :param hint: hint for human.
    :param aria_label: html input aria-label value.
    :param value: current value.
    :param error: error of the value if any.
    :param children_types: childrens types list.
    :param removable: display a button to remove the widget for optional fields.
    :param token: token used to get unique id on the form.
    """

    name: str
    "variable name, nested variables have dots."
    id: str = Field(default="")
    "variable name, nested variables have dots."
    value: T | None = Field(default=None)
    """Value of the field."""
    title: str = Field(default="")
    "Human title for the widget."
    hint: str | None = Field(default=None)
    "A help message for the the widget."

    error: str | None = Field(default=None)
    "Error message."

    aria_label: str | None = Field(default=None)
    "Non visible text alternative."
    token: str = Field(default="")
    "Unique token to ensure id are unique in the DOM."
    removable: bool = Field(default=False)
    "Indicate that the widget is removable from the dom."

    @model_validator(mode="after")
    def fill_props(self) -> Self:
        self.title = self.title or self.name.split(".")[-1]
        self.token = self.token or secrets.token_urlsafe(4).replace("_", "-")
        self.id = self.id or f"{self.name}-{self.token}".replace("_", "-").replace(
            ".", "-"
        )
        return self

    def to_html(self, renderer: AbstractTemplateRenderer) -> Markup:
        """Return the html version."""
        return Markup(renderer.render_template(self))


def _get_fullname(typ: type[Any]) -> str:
    if is_union(typ):
        typs = [_get_fullname(t) for t in typ.__args__]  # type: ignore
        return "|".join(typs)  # type: ignore
    return f"{typ.__module__}:{typ.__name__}"


TWidget = TypeVar("TWidget", bound=Widget[Any])


class CustomWidget(Generic[TWidget]):
    typ: type[Any]

    def __init__(self, typ: type[TWidget]) -> None:
        self.typ = typ


class TypeWrapper:
    """
    Wrap children types for union type.

    :param typ: Wrapped type.
    :param route_prefix: route prefix used for ajax query to build type.
    :param name: name of the field wrapped.
    :param token: unique token to render unique id.
    :param title: title to display.

    """

    def __init__(
        self,
        typ: type[Any],
        route_prefix: str,
        name: str,
        token: str,
        title: str | None = None,
    ):
        self.typ = typ
        self.route_prefix = route_prefix
        self.name = name
        self.title = title or get_title(typ)
        self.token = token

    @property
    def fullname(self) -> str:
        """Full name for the type."""
        return _get_fullname(self.typ)

    @property
    def id(self) -> str:
        """Unique id to inject in the DOM."""
        name = self.name.replace("_", "-").replace(".", "-").replace(":", "-")
        typ = self.typ.__name__.replace("_", "-")
        return f"{name}-{typ}-{self.token}"

    @property
    def params(self) -> Mapping[str, str]:
        """Params for the widget to render."""
        return {"name": self.name, "token": self.token, "title": self.title}

    @property
    def url(self) -> str:
        """Url to fetch the widget."""
        ret = f"{self.route_prefix}/pydantic-form/widgets/{self.fullname}?format=xcomponent"
        return ret



"""
Widget for field of type Set.
"""

from collections.abc import Sequence
from typing import Self

from pydantic import BaseModel, Field, model_validator

from .base import Widget


class Checkable(BaseModel):
    """A checkable field from a checklist."""

    label: str
    name: str
    value: str
    token: str
    checked: bool
    error: str | None = Field(default=None)

    id: str | None = Field(default=None)
    field_name: str | None = Field(default=None)

    @model_validator(mode="after")
    def fill_props(self) -> Self:
        self.id = f"{self.name}-{self.value}-{self.token}".replace(".", "-")
        self.field_name = f"{self.name}[]"
        return self


class ChecklistWidget(Widget[Sequence[Checkable]]):
    """
    Widget for field of type Set.
    """

    template = """
    <pydantic_form.Widget :widget_id="id" :removable="removable">
      <div class="pt-4">
        <Details>
          <Summary :id="id + '-summary'">
            <H3 :class="H3_SUMMARY_CLASS">{{ gettext(title) }}</H3>
            <pydantic_form.Error :text="error" />
          </Summary>
          <div>
            {% for value in value %}
            <div class="flex items-center mb-4">
                <Checkbox :name="value.field_name" type="checkbox"
                    :id="value.id"
                    :value="value.value"
                    :checked="value.checked" />
                <Label :for="value.id"
                    class="ms-2 text-base text-neutral-900 dark:text-white">
                    {{- gettext(value.label) -}}
                </Label>
                <pydantic_form.Error :text="value.error" />
            </div>
            {% endfor %}
          </div>
        </Details>
      </div>
    </pydantic_form.Widget>
    """



"""Text components."""

from collections.abc import Mapping

from xcomponent import XNode

from fastlife.adapters.xcomponent.catalog import catalog


@catalog.component
def P(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    <p> html tag rendering

    :param children: child node.
    :param globals: injected root component globals variables.
    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.P_CLASS`.
    """
    return """
    <p id={id} class={class_ or globals.P_CLASS}>
      {children}
    </p>
    """



"""Pydantic models"""

from collections.abc import Sequence

from markupsafe import Markup
from pydantic import Field

from fastlife.service.templates import AbstractTemplateRenderer

from .base import TWidget, Widget


class ModelWidget(Widget[Sequence[TWidget]]):
    template = """
    <pydantic_form.Widget :widget_id="id" :removable="removable">
    <div id="{{id}}"{% if nested %} class="m-4"{%endif%}>
        {% if nested %}
        <Details>
        <Summary :id="id + '-summary'">
            <H3 :class="H3_SUMMARY_CLASS">{{ gettext(title) }}</H3>
            <pydantic_form.Error :text="error" />
        </Summary>
        <div>
            {% for child in children_widgets %}
            {{ child }}
            {% endfor %}
        </div>
        </Details>
        {% else %}
            {% for child in children_widgets %}
            {{ child }}
            {% endfor %}
        {% endif %}
    </div>
    </pydantic_form.Widget>
    """

    nested: bool = Field(default=False)
    children_widgets: list[str] | None = Field(default=None)

    def to_html(self, renderer: AbstractTemplateRenderer) -> Markup:
        """Return the html version."""
        self.children_widgets = [child.to_html(renderer) for child in self.value or []]
        return Markup(renderer.render_template(self))



"""
Widget for field of type Enum or Literal.
"""

from collections.abc import Sequence

from pydantic import Field, field_validator
from typing_extensions import TypedDict

from fastlife.adapters.xcomponent.catalog import catalog

from .base import Widget

OptionItem = str | int


class Option(TypedDict):
    value: OptionItem
    text: OptionItem


@catalog.component
def DropDownWidgetOption(id: str, opt: Option) -> str:
    return """
    <Option
        value={opt.value}
        id={id + "-" + opt.value.replace(" ", " -")}
        selected={value and value == opt.value}
        >
      {globals.gettext(opt.text)}
    </Option>
    """


class DropDownWidget(Widget[str]):
    """
    Widget for field of type Enum or Literal.
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div class="pt-4">
        <Label for={id}>{ globals.gettext(title) }</Label>
        <Select name={name} id={id}>
          {
            for opt in options  {
              <DropDownWidgetOption id={id} opt={opt} value={value}/>
            }
          }
        </Select>
        <OptionalErrorText text={error} />
        <Hint text={hint} />
      </div>
    </Widget>
    """

    options: list[Option] = Field(default_factory=list)

    @field_validator("options", mode="before")
    @classmethod
    def validate_options(
        cls, options: Sequence[OptionItem | tuple[OptionItem, OptionItem]] | None
    ) -> Sequence[Option]:
        if not options:
            return []
        ret: list[Option] = []
        for opt in options:
            if isinstance(opt, tuple):
                ret.append({"value": opt[0], "text": opt[1]})
            else:
                ret.append({"value": opt, "text": opt})
        return ret



"""
Adapters based on top of FastAPI.

At the momdent, the code is highly coupled with FastAPI,
some part of the code import fastapi directly for legacy reason,
but it shouldn't.

All FastAPI dependency should be contained in this module.
"""



"""HTTP Route."""

from collections.abc import Callable, Coroutine
from typing import TYPE_CHECKING, Any

from fastapi.routing import APIRoute

from fastlife.domain.model.asgi import ASGIRequest, ASGIResponse

if TYPE_CHECKING:
    from fastlife.service.registry import DefaultRegistry  # coverage: ignore


class Route(APIRoute):
    """
    Routing for fastlife application.

    The fastlife router construct fastlife request object in order to
    have the registry property available in every received request.
    """

    _registry: "DefaultRegistry"
    """
    The application registry.

    this static variable is initialized by the configurator during
    the startup and keep the registry during the lifetime of the application.

    this variable should be accessed via the request object or the
    {class}`fastlife.config.Registry` depenency injection.
    """

    def get_route_handler(
        self,
    ) -> Callable[[ASGIRequest], Coroutine[Any, Any, ASGIResponse]]:
        """
        Replace the request object by the fastlife request associated with the registry.
        """
        orig_route_handler = super().get_route_handler()

        async def route_handler(request: ASGIRequest) -> ASGIResponse:
            req = self._registry.request_factory(request)
            return await orig_route_handler(req)

        return route_handler



"""
Template rending based on JinjaX.
"""

import logging
import textwrap
from collections.abc import Sequence
from typing import (
    TYPE_CHECKING,
    Any,
)

from markupsafe import Markup
from pydantic.fields import FieldInfo

from fastlife import Request
from fastlife.adapters.fastapi.form import FormModel
from fastlife.adapters.fastapi.localizer import get_localizer
from fastlife.adapters.jinjax.widget_factory.factory import WidgetFactory
from fastlife.config import Configurator, configure
from fastlife.domain.model.template import InlineTemplate

if TYPE_CHECKING:
    from fastlife.settings import Settings  # coverage: ignore

from fastlife.service.templates import (
    AbstractTemplateRenderer,
    AbstractTemplateRendererFactory,
)
from fastlife.shared_utils.resolver import resolve, resolve_path

from .jinjax_ext.inspectable_catalog import InspectableCatalog

log = logging.getLogger(__name__)


def build_searchpath(template_search_path: str) -> Sequence[str]:
    """
    Build the path containing templates.

    Path may be absolute directories or directories relative to a python
    package. For instance, the `fastlife:components` is the directory components
    inside the fastlife installation dir.

    :param template_search_path: list of path separated by a comma (`,`).
    :return: List resolved path.
    """
    searchpath: list[str] = []
    paths = template_search_path.split(",")

    for path in paths:
        if ":" in path:
            searchpath.append(resolve_path(path))
        else:
            searchpath.append(path)
    return searchpath


class JinjaxRenderer(AbstractTemplateRenderer):
    """Render templates using JinjaX."""

    def __init__(
        self,
        catalog: InspectableCatalog,
        request: Request,
    ):
        super().__init__(request)
        self.catalog = catalog
        self.settings = request.registry.settings
        self.translations = get_localizer(request)
        self.globals["pydantic_form"] = self.pydantic_form

    def render_template(self, template: InlineTemplate) -> str:
        """
        Render the JinjaX component with the given parameter.

        :param template: the template to render
        :param globals: parameters that will be used by the JinjaX component and all its
            child components without "props drilling".
        :param params: parameters used to render the template.
        """
        params = template.model_dump()
        src = (
            f"{{# def {', '.join(params.keys())} #}}\n"
            f"{textwrap.dedent(template.template)}"
        )
        return self.catalog.render(
            template.__class__.__qualname__,
            __source=src,
            __globals=self.globals,
            **params,
        )

    def pydantic_form(
        self, model: FormModel[Any], *, token: str | None = None
    ) -> Markup:
        """
        Generate HTML markup to build a form from the given form model.

        :param model: the form model that will be transformed to markup.
        :param token: a token used to ensure that unique identifier are unique.
        :return: HTML Markup generated by composign fields widgets.
        """
        return WidgetFactory(self, token).get_markup(model)

    def pydantic_form_field(
        self,
        model: type[Any],
        *,
        name: str | None,
        token: str | None,
        removable: bool,
        field: FieldInfo | None,
    ) -> Markup:
        """
        Generate HTML for a particular field in a form.

        This function is used to generate union subtypes in Ajax requests.
        :param model: a pydantic or python builtin type that is requests to be rendered
        :param name: name for the field
        :param token: the token of the form to render unique identifiers
        :param removable: add a way let the user remove the widget
        :param field: only render this particular field for the model.
        :return: HTML Markup.
        """
        return (
            WidgetFactory(self, token)
            .get_widget(
                model,
                form_data={},
                form_errors={},
                prefix=(name or self.request.registry.settings.form_data_model_prefix),
                removable=removable,
                field=field,
            )
            .to_html(self)
        )


class JinjaxEngine(AbstractTemplateRendererFactory):
    """
    The default template renderer factory. Based on JinjaX.

    :param settings: setting used to configure jinjax.
    """

    def __init__(self, settings: "Settings") -> None:
        globals = resolve(settings.jinjax_global_catalog_class)().model_dump()

        self.catalog = InspectableCatalog(
            use_cache=settings.jinjax_use_cache,
            file_ext=settings.jinjax_file_ext,
            auto_reload=settings.jinjax_auto_reload,
            globals=globals,
        )
        self.catalog.jinja_env.add_extension("jinja2.ext.i18n")
        for path in build_searchpath(settings.template_search_path):
            self.catalog.add_folder(path)

    def __call__(self, request: Request) -> AbstractTemplateRenderer:
        """Build the renderer to render request for template."""
        return JinjaxRenderer(
            self.catalog,
            request,
        )


@configure
def includeme(conf: Configurator) -> None:
    conf.add_renderer(
        conf.registry.settings.jinjax_file_ext, JinjaxEngine(conf.registry.settings)
    )



"""HTTP Form serialization."""

from collections.abc import Callable, Mapping
from typing import Any

from fastapi import Depends

from fastlife.adapters.fastapi.form_data import MappingFormData
from fastlife.adapters.fastapi.request import Registry
from fastlife.domain.model.form import FormModel, T


def form_model(
    cls: type[T], name: str | None = None
) -> Callable[[Mapping[str, Any]], FormModel[T]]:
    """
    Build a model, a class of type T based on Pydandic Base Model from a form payload.
    """

    def to_model(data: MappingFormData, registry: Registry) -> FormModel[T]:
        prefix = name or registry.settings.form_data_model_prefix
        if not data:
            return FormModel[T].default(prefix, cls)
        return FormModel[T].from_payload(prefix, cls, data)

    return Depends(to_model)



"""Handle Pydantic SecretStr type."""

from collections.abc import Mapping
from typing import Any

from pydantic import SecretStr
from pydantic.fields import FieldInfo

from fastlife.adapters.xcomponent.pydantic_form.widget_factory.base import (
    BaseWidgetBuilder,
)
from fastlife.adapters.xcomponent.pydantic_form.widgets.base import Widget
from fastlife.adapters.xcomponent.pydantic_form.widgets.text import PasswordWidget


class SecretStrBuilder(BaseWidgetBuilder[SecretStr]):
    """Builder for Pydantic SecretStr."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for Pydantic SecretStr."""
        return issubclass(typ, SecretStr)

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],
        field: FieldInfo | None,
        value: SecretStr | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> Widget[SecretStr]:
        """Build the widget."""
        return PasswordWidget(
            name=field_name,
            placeholder=str(field.examples[0]) if field and field.examples else None,
            removable=removable,
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            error=form_errors.get(field_name),
            new_password="new-password" in field.metadata if field else False,
        )



"""Models relative to the security."""

import secrets

from pydantic import BaseModel


def create_csrf_token() -> str:
    """A helper that create a csrf token."""
    return secrets.token_urlsafe(5)


class CSRFToken(BaseModel):
    """Represent the CSRF Token"""

    name: str
    """Name of the token while serialized."""
    value: str
    """Value that must match between parts, cookie and posted form."""



from collections.abc import Sequence
from typing import Any

from markupsafe import Markup
from pydantic import Field

from fastlife.service.templates import AbstractTemplateRenderer

from .base import TWidget, TypeWrapper, Widget


class SequenceWidget(Widget[Sequence[TWidget]]):
    template = """
    <Widget widget_id={id} removable={removable}>
      <Details id={id}>
        <Summary id={id + '-summary'}>
          <H3 class={globals.H3_SUMMARY_CLASS}>{ globals.gettext(title) }</H3>
          <OptionalErrorText text={error} />
        </Summary>
        <div>
          <script>
            // this function should be added once.
            function getName(name, id) {
              const el = document.getElementById(id + '-content');
              const len = el.dataset.length;
              el.dataset.length = parseInt(len) + 1;
              return name + '.' + len;
            }
          </script>
          <div id={id + "-content"} class="m-4"
            data-length={len(children_widgets)}>
            { let container_id = id + "-children-container" }
            <div id={container_id}>
              {
                for child in children_widgets {
                  child
                }
              }
            </div>
          </div>
          <div>
            { let container_id = "#" + id + "-children-container" }
            { let add_id = id + "-add" }
            { let vals = 'js:{"name": getName("' + wrapped_type.name + '", "' + id + '")'
                + ', "token": "'
                + wrapped_type.token + '", "removable": true}'
            }
            <Button type="button"
                hx-target={container_id} hx-swap="beforeend"
                id={add_id}
                hx-vals={vals}
                hx-get={wrapped_type.url}>
              Add
            </Button>
          </div>
        </div>
      </Details>
    </Widget>
    """

    item_type: type[Any]
    wrapped_type: TypeWrapper | None = Field(default=None)
    children_widgets: list[str] | None = Field(default=None)

    def build_item_type(self, route_prefix: str) -> TypeWrapper:
        return TypeWrapper(self.item_type, route_prefix, self.name, self.token)

    def to_html(self, renderer: "AbstractTemplateRenderer") -> Markup:
        """Return the html version."""
        self.wrapped_type = self.build_item_type(renderer.route_prefix)
        self.children_widgets = [
            Markup(item.to_html(renderer)) for item in self.value or []
        ]
        return Markup(renderer.render_template(self))



"""Use ``<Icon>`` to load icons."""

import html
import re
from functools import cache
from importlib import util
from pathlib import Path
from typing import Literal
from zipfile import ZipFile

from fastlife.adapters.xcomponent.catalog import catalog


@cache
def icon_path() -> Path:
    spec = util.find_spec("heroicons")
    if spec is None or spec.origin is None:
        raise ValueError("Install heroicons first")  # coverage: ignore
    return Path(spec.origin).parent


@cache
def get_from_zip(
    name: str,
    mode: str,
) -> str | None:
    iconzip = icon_path() / "heroicons.zip"
    with ZipFile(iconzip, "r") as zip_ref:
        file_list = zip_ref.namelist()
        for filepath in file_list:
            if filepath == f"{mode}/{name}.svg":
                with zip_ref.open(filepath) as file:
                    return file.read().decode("utf-8")
    return None


@catalog.function
def load_icon(
    name: str,
    mode: str,
    id: str | None = None,
    title: str | None = None,
    class_: str | None = None,
) -> str:
    """
    Function to load the icon behind the scene for  the ``<Icon>`` component.

    :param name: name of the icon.
    :param mode: mode of the hero icon.
    :param id: dom unique identifier.
    :param title: title that can be bubble to the icon.
    :param class_: css class to set.
    """
    icon = get_from_zip(name, mode)
    if not icon:
        return ""

    attrs = ""
    if id:
        attrs += f' id="{html.escape(id)}"'
    if class_:
        attrs += f' class="{html.escape(class_)}"'
    if attrs:
        icon = icon.replace(' viewBox="', f'{attrs} viewBox="', 1)
    if title:
        icon = re.sub(
            r"(<path[^>]*?)/>",
            rf"\1><title>{html.escape(title)}</title></path>",
            icon,
        )
    return icon


@catalog.component
def Icon(
    name: str,
    mode: Literal["micro", "mini", "outline", "solid"] = "solid",
    id: str | None = None,
    title: str | None = None,
    class_: str | None = None,
) -> str:
    """
    Add an icon from hero icon package. The svg is already injected to the DOM.

    :param name: name of the icon.
    :param mode: mode of the hero icon.
    :param id: dom unique identifier.
    :param title: title that can be bubble to the icon.
    :param class_: css class to set.
    """
    return """
    <>{load_icon(name, mode, id, title, class_)}</>
    """



"""HTTP Request representation in a python object."""

from typing import TYPE_CHECKING, Any, Generic

from fastlife.domain.model.asgi import ASGIRequest
from fastlife.domain.model.csrf import CSRFToken, create_csrf_token
from fastlife.domain.model.security_policy import TClaimedIdentity, TIdentity
from fastlife.service.registry import TRegistry

if TYPE_CHECKING:
    from fastlife.service.security_policy import (  # coverage: ignore
        AbstractSecurityPolicy,
        HasPermission,
    )


class GenericRequest(ASGIRequest, Generic[TRegistry, TIdentity, TClaimedIdentity]):
    """HTTP Request representation."""

    registry: TRegistry
    """Direct access to the application registry."""
    locale_name: str
    """Request locale used for the i18n of the response."""

    security_policy: (
        "AbstractSecurityPolicy[TRegistry, TIdentity, TClaimedIdentity] | None"
    )
    """Request locale used for the i18n of the response."""

    renderer_globals: dict[str, Any]

    def __init__(self, registry: TRegistry, request: ASGIRequest) -> None:
        super().__init__(request.scope, request.receive)
        self.registry = registry
        self.locale_name = registry.locale_negociator(self)
        self.security_policy = None  # build it from the ? registry
        self.renderer_globals = {}
        self._csrf_token: CSRFToken | None = None

    @property
    def csrf_token(self) -> CSRFToken:
        if self._csrf_token is None:
            name = self.registry.settings.csrf_token_name
            value = self.cookies.get(name) or create_csrf_token()
            self._csrf_token = CSRFToken(name=name, value=value)
        return self._csrf_token

    def add_renderer_globals(self, **kwargs: Any) -> None:
        """
        Add global variables to the template renderer context for the current request.
        """
        self.renderer_globals.update(kwargs)

    async def has_permission(
        self, permission: str
    ) -> "HasPermission | type[HasPermission]":
        """
        A helper to check that a user has the given permission.

        Not that this method does not raised, it return a boolean like object.
        It allows batch permission checks.
        You might need to check multiple permissions in different contexts or
        for different resources before raising an http error.
        """
        if self.security_policy is None:
            raise RuntimeError(
                f"Request {self.url.path} require a security policy, "
                "explicit fastlife.service.security_policy.InsecurePolicy is required."
            )

        return await self.security_policy.has_permission(permission)

    def url_path_for(self, name: str, /, **path_params: Any) -> str:
        """
        Return the url pathinfo for the given route and route parameters.

        :param name: the name of the route
        :param path_params: parameters for the route.
        """
        url_path_provider: Any = self.scope.get("router") or self.scope.get("app")
        return url_path_provider.url_path_for(name, **path_params)



"""
Template renderer based on XComponent.

More template engine can be registered using the configurator method
{meth}`add_renderer <fastlife.config.configurator.GenericConfigurator.add_renderer>`
"""

from typing import Any

from markupsafe import Markup
from pydantic.fields import FieldInfo

from fastlife import Request
from fastlife.adapters.xcomponent.pydantic_form.widget_factory.factory import (
    WidgetFactory,
)
from fastlife.config import Configurator, configure
from fastlife.domain.model.form import FormModel
from fastlife.domain.model.template import InlineTemplate
from fastlife.service.templates import (
    AbstractTemplateRenderer,
    AbstractTemplateRendererFactory,
)
from fastlife.settings import Settings
from fastlife.shared_utils.resolver import resolve

from .catalog import catalog


class XTemplateRenderer(AbstractTemplateRenderer):
    """
    An object that will be initialized by an AbstractTemplateRendererFactory,
    passing the request to process.
    """

    request: Request
    """Associated request that needs a response."""

    def __init__(self, globals: dict[str, Any], request: Request) -> None:
        self.request = request
        self.globals: dict[str, Any] = {**globals}
        self.globals["pydantic_form"] = self.pydantic_form

    def pydantic_form(
        self, model: FormModel[Any], *, token: str | None = None
    ) -> Markup:
        """
        Generate HTML markup to build a form from the given form model.

        :param model: the form model that will be transformed to markup.
        :param token: a token used to ensure that unique identifier are unique.
        :return: HTML Markup generated by composign fields widgets.
        """
        return WidgetFactory(self, token).get_markup(model)

    def pydantic_form_field(
        self,
        model: type[Any],
        *,
        name: str | None,
        token: str | None,
        removable: bool,
        field: FieldInfo | None,
    ) -> Markup:
        """
        Generate HTML for a particular field in a form.

        This function is used to generate union subtypes in Ajax requests.
        :param model: a pydantic or python builtin type that is requests to be rendered
        :param name: name for the field
        :param token: the token of the form to render unique identifiers
        :param removable: add a way let the user remove the widget
        :param field: only render this particular field for the model.
        :return: HTML Markup.
        """
        return (
            WidgetFactory(self, token)
            .get_widget(
                model,
                form_data={},
                form_errors={},
                prefix=(name or self.request.registry.settings.form_data_model_prefix),
                removable=removable,
                field=field,
            )
            .to_html(self)
        )

    @property
    def route_prefix(self) -> str:
        """Used to buid pydantic form widget that do ajax requests."""
        return self.request.registry.settings.fastlife_route_prefix

    def render_template(self, template: InlineTemplate) -> str:
        """
        Render an inline template.

        :param template: the template to render.
        :return: The template rendering result.
        """
        params = template.model_dump()
        return catalog.render(
            template.template,
            globals=self.globals,
            **params,
        )


class XRendererFactory(AbstractTemplateRendererFactory):
    """
    The template render factory.
    """

    def __init__(self, settings: "Settings") -> None:
        self.globals = resolve(settings.jinjax_global_catalog_class)().model_dump()

    def __call__(self, request: Request) -> AbstractTemplateRenderer:
        """
        While processing an HTTP Request, a renderer object is created giving
        isolated context per request.

        :param Request: the HTTP Request to process.
        :return: The renderer object that will process that request.
        """
        return XTemplateRenderer(globals=self.globals, request=request)


@configure
def includeme(conf: Configurator) -> None:
    conf.add_renderer("xcomponent", XRendererFactory(conf.registry.settings))



"""
Widget for field of type Set.
"""

from collections.abc import Sequence
from typing import Any, Self

from pydantic import BaseModel, Field, model_validator

from fastlife.adapters.xcomponent.catalog import catalog

from .base import Widget


class Checkable(BaseModel):
    """A checkable field from a checklist."""

    label: str
    name: str
    value: str
    token: str
    checked: bool
    error: str | None = Field(default=None)

    id: str | None = Field(default=None)
    field_name: str | None = Field(default=None)

    @model_validator(mode="after")
    def fill_props(self) -> Self:
        self.id = f"{self.name}-{self.value}-{self.token}".replace(".", "-")
        self.field_name = f"{self.name}[]"
        return self


@catalog.component
def PydanticFormChecklistItem(value: Checkable, globals: Any) -> str:
    return """
    <div class="flex items-center mb-4">
        <Checkbox name={value.field_name} type="checkbox"
            id={value.id}
            value={value.value}
            checked={value.checked} />
        <Label for={value.id}
            class="ms-2 text-base text-neutral-900 dark:text-white">
            {globals.gettext(value.label)}
        </Label>
        <OptionalErrorText text={value.error} />
    </div>
    """


class ChecklistWidget(Widget[Sequence[Checkable]]):
    """
    Widget for field of type Set.
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div class="pt-4">
        <Details>
          <Summary id={id + '-summary'}>
            <H3 class={globals.H3_SUMMARY_CLASS}>{globals.gettext(title)}</H3>
            <OptionalErrorText text={error} />
          </Summary>
          <div>
            {
              for v in value {
                <PydanticFormChecklistItem value={v} />
              }
            }
          </div>
        </Details>
      </div>
    </Widget>
    """



"""Build you own middleware."""

import abc

from fastlife.domain.model.asgi import Receive, Scope, Send


class AbstractMiddleware(abc.ABC):
    """
    Abstract Base Class that represent a fastlife middleware.

    Starlette provide a middleware stack but does not have an abstract base class.
    This is the only reason this class exists.

    Fastlife middleware are starlette middlewares.
    """

    @abc.abstractmethod
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Called every time an http request is reveived.

        This method before the request object even exists.
        """



"""
Template rending based on JinjaX.
"""

import ast
import textwrap
from typing import Any


def generate_docstring(
    func_def: ast.FunctionDef, component_name: str, add_content: bool
) -> str:
    """Generate a docstring for a JinjaX component."""
    # Extract function name and docstring
    docstring = (ast.get_docstring(func_def, clean=True) or "").strip()
    if docstring:
        docstring = textwrap.dedent(docstring)
        docstring_lines = [line for line in docstring.split("\n")]
        # Add a newline for separation after the function docstring
        docstring_lines.append("")
    else:
        docstring_lines = []

    component_params: list[str] = []

    # Function for processing an argument and adding its docstring lines
    def process_arg(arg: ast.arg, default_value: Any = None) -> None:
        arg_name = arg.arg
        param_desc = ""
        # Extract the type annotation (if any)
        if (
            isinstance(arg.annotation, ast.Subscript)
            and isinstance(arg.annotation.value, ast.Name)
            and arg.annotation.value.id == "Annotated"
        ):
            # For Annotated types, we expect the first argument to be the type and
            # the second to be the description
            type_annotation = arg.annotation.slice.elts[0]  # type: ignore
            param_type = ast.unparse(type_annotation)  # type: ignore

            if len(arg.annotation.slice.elts) > 1 and isinstance(  # type: ignore
                arg.annotation.slice.elts[1],  # type: ignore
                ast.Constant,  # type: ignore
            ):
                param_desc = arg.annotation.slice.elts[1].value  # type: ignore
        else:
            # Otherwise, just use the type if available
            param_type = ast.unparse(arg.annotation) if arg.annotation else "Any"

        # Build the parameter docstring line
        docstring_lines.append(f":param {arg_name.rstrip('_')}: {param_desc}".strip())

        # Build the string representation of the parameter
        param_str = f"{arg_name}: {param_type}"
        if default_value is not None:
            param_str += f" = {ast.unparse(default_value)}"  # type: ignore

        component_params.append(param_str)

    # Process keyword-only arguments
    kwonlyargs = func_def.args.kwonlyargs
    kw_defaults = func_def.args.kw_defaults

    for arg, default in zip(kwonlyargs, kw_defaults, strict=False):
        process_arg(arg, default)

    if add_content:
        component_params.append("content: Any")
        docstring_lines.append(":param content: child node.")

    return (
        f"{component_name}({', '.join(component_params)})"
        + "\n"
        + "\n    "
        + ("\n    ".join(docstring_lines).strip()).replace("\n    \n", "\n\n")
        + "\n"
    )



"""
Base class to of the template renderer.

Fastlife comes with two rendering engines:
{class}`fastlife.adapters.jinjax.renderer.JinjaxEngine`,
and {class}`fastlife.adapters.xcomponent.renderer.XRendererFactory`,
they have to be installed using python packaging extra syntax for your tool,
fastlifeweb[jinjax] or fastlifeweb[xcomponent] since those engine are not
installed by default.

More template engine can be registered using the configurator method
{meth}`add_renderer <fastlife.config.configurator.GenericConfigurator.add_renderer>`
"""

import abc
from collections.abc import Mapping
from typing import Any

from markupsafe import Markup
from pydantic.fields import FieldInfo

from fastlife import Request, Response
from fastlife.adapters.fastapi.form import FormModel
from fastlife.domain.model.template import InlineTemplate


class AbstractTemplateRenderer(abc.ABC):
    """
    An object that will be initialized by an AbstractTemplateRendererFactory,
    passing the request to process.
    """

    request: Request
    """Associated request that needs a response."""

    def __init__(self, request: Request) -> None:
        self.request = request
        self.globals: dict[str, Any] = {}

    @property
    def route_prefix(self) -> str:
        """Used to buid pydantic form widget that do ajax requests."""
        return self.request.registry.settings.fastlife_route_prefix

    def render(
        self,
        template: str,
        *,
        status_code: int = 200,
        content_type: str = "text/html",
        globals: Mapping[str, Any] | None = None,
        params: InlineTemplate,
    ) -> Response:
        """
        Render the template and build the HTTP Response.
        """
        request = self.request
        if globals:
            self.globals.update(globals)
        data = self.render_template(params)
        resp = Response(
            data, status_code=status_code, headers={"Content-Type": content_type}
        )
        resp.set_cookie(
            request.csrf_token.name,
            request.csrf_token.value,
            httponly=True,
            secure=request.url.scheme == "https",
            samesite="strict",
            max_age=60 * 15,
        )
        return resp

    @abc.abstractmethod
    def render_template(self, template: InlineTemplate) -> str:
        """
        Render an inline template.

        :param template: the template to render.
        :return: The template rendering result.
        """

    @abc.abstractmethod
    def pydantic_form(
        self, model: FormModel[Any], *, token: str | None = None
    ) -> Markup:
        """
        Generate HTML markup to build a form from the given form model.

        :param model: the form model that will be transformed to markup.
        :param token: a token used to ensure that unique identifier are unique.
        :return: HTML Markup generated by composign fields widgets.
        """

    @abc.abstractmethod
    def pydantic_form_field(
        self,
        model: type[Any],
        *,
        name: str | None,
        token: str | None,
        removable: bool,
        field: FieldInfo | None,
    ) -> Markup:
        """
        Generate HTML for a particular field in a form.

        This function is used to generate union subtypes in Ajax requests.
        :param model: a pydantic or python builtin type that is requests to be rendered
        :param name: name for the field
        :param token: the token of the form to render unique identifiers
        :param removable: add a way let the user remove the widget
        :param field: only render this particular field for the model.
        :return: HTML Markup.
        """


class AbstractTemplateRendererFactory(abc.ABC):
    """
    The template render factory.
    """

    @abc.abstractmethod
    def __call__(self, request: Request) -> AbstractTemplateRenderer:
        """
        While processing an HTTP Request, a renderer object is created giving
        isolated context per request.

        :param Request: the HTTP Request to process.
        :return: The renderer object that will process that request.
        """



import ast
import re
from pathlib import Path
from typing import Any, ClassVar, cast

from docutils import nodes
from sphinx.addnodes import index, pending_xref
from sphinx.application import Sphinx
from sphinx.builders import Builder
from sphinx.directives import ObjectDescription
from sphinx.domains import Domain, ObjType
from sphinx.environment import BuildEnvironment
from sphinx.roles import XRefRole
from sphinx.util import relative_uri  # type: ignore

from fastlife.adapters.jinjax.renderer import JinjaxEngine
from fastlife.settings import Settings


def create_ref_node(arg_type: str) -> nodes.Node:
    """
    Creates a cross-reference node if a type has been inferred.

    At the monent, type inferred are types containing a `.`, otherwise
    considered as a simple type from the python standard library.

    This function is used to create links in component parameters.

    :param arg_type: type to render.
    """
    if "|" in arg_type:
        typs = [t.strip() for t in arg_type.split("|")]
        typs_group = nodes.literal()
        for idx, typ in enumerate(typs):
            if idx:
                typs_group += nodes.Text(" | ")
            typs_group += create_ref_node(typ)
        return typs_group

    if "." in arg_type:
        ref_node = pending_xref(
            "",
            refdomain="py",
            reftype="class",
            reftarget=arg_type,
            refexplicit=True,
            refwarn=True,
            classes=[],
        )
        ref_node += nodes.Text(arg_type)
        return ref_node
    else:
        return nodes.inline(text=arg_type, classes=["jinjax-type"])


def handle_arg_type(arg_type: str, signature_node: nodes.literal) -> None:
    """Add arg_type to signature_node."""
    complex_type_pattern = r"(\w+)\[(.+)\]"

    match = re.match(complex_type_pattern, arg_type)
    if match:
        outer_type, inner_type = match.groups()
        signature_node += nodes.inline(text=f"{outer_type}[", classes=["jinjax-type"])
        signature_node += create_ref_node(inner_type)
        signature_node += nodes.inline(text="]", classes=["jinjax-type"])
    else:
        signature_node += create_ref_node(arg_type)


class JinjaxComponent(ObjectDescription[str]):
    """Description of a Jinjax component."""

    def run(self) -> list[nodes.Node]:
        """Generate structured and styled documentation for the directive."""
        container_node = nodes.container(classes=["jinjax-component"])

        # Extract the signature (first line of arguments)
        signature_text = self.arguments[0]
        component_name, rest = signature_text.split("(", 1)

        params = rest.rstrip(")")
        if params:
            signature = f"def component(*, {params}): pass"
        else:
            # no parameters in the component
            signature = "def component(): pass"

        astree = ast.parse(signature)
        func_def = cast(ast.FunctionDef, astree.body[0])

        # Create a colorized signature line with separate spans for each part
        signature_node = nodes.literal(classes=["jinjax-signature"])
        signature_node += nodes.inline(text="<")
        signature_node += nodes.inline(
            text=component_name, classes=["jinjax-component-name"]
        )

        def process_arg(arg: ast.arg, default_value: Any, signature_node: Any) -> None:
            arg_name = arg.arg.replace("_", "-").rstrip("-")
            arg_type = ast.unparse(arg.annotation) if arg.annotation else "Any"
            signature_node += nodes.inline(text=" ")

            signature_node += nodes.inline(text=arg_name, classes=["jinjax-arg"])
            signature_node += nodes.inline(text=": ")

            handle_arg_type(arg_type, signature_node)

            if default_value is not None:
                signature_node += nodes.inline(text=" = ")
                signature_node += nodes.inline(
                    text=ast.unparse(default_value), classes=["jinjax-default"]
                )

        # Process keyword-only arguments
        kwonlyargs = func_def.args.kwonlyargs
        kw_defaults = func_def.args.kw_defaults

        has_content = False
        for arg, default in zip(kwonlyargs, kw_defaults, strict=False):
            if arg.arg == "content":
                has_content = True
            else:
                process_arg(arg, default, signature_node)

        signature_close_wrapper = nodes.container(classes=["jinjax-signature-close"])
        if has_content:
            signature_node += nodes.inline(text=">")
            signature_node += nodes.inline(
                text="{{- content -}}", classes=["jinjax-child-content"]
            )
            signature_close_wrapper += nodes.inline(text="</")
            signature_close_wrapper += nodes.inline(
                text=component_name, classes=["jinjax-component-name"]
            )
            signature_close_wrapper += nodes.inline(text=">")
        else:
            signature_node += nodes.inline(text=" />")
            signature_close_wrapper += nodes.inline(text="")
        signature_node += signature_close_wrapper

        signature_node_wrapper = nodes.container(classes=["jinjax-signature-wrapper"])
        signature_node_wrapper += signature_node

        container_node += nodes.paragraph("", "", signature_node_wrapper)

        if self.content:
            content = self.parse_content_to_nodes(allow_section_headings=True)
            for content_node in content:
                container_node += content_node

        index_entry = index(
            entries=[
                (
                    "single",
                    f"{component_name} (JinjaX component)",
                    f"jinjax-component-{component_name}",
                    "",
                    None,
                )
            ]
        )
        return [index_entry, container_node]


class JinjaxDomain(Domain):
    """Custom domain for Jinjax components."""

    name = "jinjax"
    label = "Jinjax"
    roles = {  # noqa: RUF012
        "component": XRefRole(),
    }
    directives = {  # noqa: RUF012
        "component": JinjaxComponent,
    }
    object_types = {  # noqa: RUF012
        "component": ObjType("component", "component"),
    }

    _components: ClassVar[list[str]] = []

    @classmethod
    def register(cls, component: str) -> None:
        cls._components.append(component)

    def resolve_xref(
        self,
        env: BuildEnvironment,
        fromdocname: str,
        builder: Builder,
        typ: str,
        target: str,
        node: pending_xref,
        contnode: nodes.Element,
    ) -> nodes.Element | None:
        """Resolve the pending_xref *node* with the given *typ* and *target*.

        This method should return a new node, to replace the xref node,
        containing the *contnode* which is the markup content of the
        cross-reference.

        If no resolution can be found, None can be returned; the xref node will
        then given to the :event:`missing-reference` event, and if that yields no
        resolution, replaced by *contnode*.

        The method can also raise :exc:`sphinx.environment.NoUri` to suppress
        the :event:`missing-reference` event being emitted.
        """
        ret = self.resolve_any_xref(env, fromdocname, builder, target, node, contnode)
        return ret[0][1] if ret else None

    def resolve_any_xref(
        self,
        env: BuildEnvironment,
        fromdocname: str,
        builder: Builder,
        target: str,
        node: pending_xref,
        contnode: nodes.Element,
    ) -> list[tuple[str, nodes.Element]]:
        """Resolve the pending_xref *node* with the given *target*.

        The reference comes from an "any" or similar role, which means that we
        don't know the type.  Otherwise, the arguments are the same as for
        {meth}`JinjaxDomain.resolve_xref`.

        The method must return a list (potentially empty) of tuples
        ``('domain:role', newnode)``, where ``'domain:role'`` is the name of a
        role that could have created the same reference, e.g. ``'py:func'``.
        ``newnode`` is what {meth}`JinjaxDomain.resolve_xref` would return.

        .. versionadded:: 1.3
        """
        if target in self._components:
            ref_uri = f"components/{target}.html"
            relative_link = relative_uri(fromdocname, ref_uri)

            if isinstance(contnode, nodes.literal) and contnode.astext() == target:
                contnode = nodes.literal("", f"<{target}/>")

            newnode = nodes.reference("", "", contnode, refuri=relative_link)
            return [("jinjax:component", newnode)]

        return []


def run_autodoc(app: Sphinx) -> None:
    """Run autodoc for a single package.

    :return: The top level module name, relative to the api directory.
    """
    renderer = JinjaxEngine(
        Settings(template_search_path=app.config.jinjax_template_search_path)
    )
    outdir = Path(app.srcdir) / app.config.jinjax_doc_output_dir
    outdir.mkdir(parents=True, exist_ok=True)
    toctree = ""

    ignores = [re.compile(rx) for rx in app.config.jinjax_ignores_pattern.split(",")]

    for component in renderer.catalog.iter_components(ignores=ignores):
        outfile = outdir / f"{component.name}.rst"

        outfile.write_text(
            f"{component.name}\n"
            f"{'=' * len(component.name)}\n"
            f"\n{component.build_docstring()}"
        )
        toctree += f"   {component.name}\n"
        JinjaxDomain.register(component.name)

    outfile = outdir / "index.rst"
    outfile.write_text(INDEX_TPL.format(toctree=toctree))


INDEX_TPL = """\

.. _jinjax-components:

JinjaX Components
=================

fastlife comes with a set of compoments:

.. toctree::
   :titlesonly:

{toctree}
"""


def setup(app: Sphinx) -> None:
    app.add_config_value("jinjax_doc_index_template", INDEX_TPL, "env")
    app.add_config_value("jinjax_doc_output_dir", "components", "env")
    app.add_config_value("jinjax_template_search_path", "fastlife:components", "env")
    app.add_config_value("jinjax_ignores_pattern", r"^icons\..*", "env")
    app.add_domain(JinjaxDomain)
    app.connect("builder-inited", run_autodoc)



from typing import Any

from xcomponent import Catalog

catalog = Catalog()
"""The catalog to register components."""


@catalog.function(name="len")
def length(iterable: Any) -> int:
    return len(iterable)



"""Configure fastlife app for dependency injection."""

from .configurator import Configurator, GenericConfigurator, configure
from .exceptions import exception_handler
from .resources import resource, resource_view
from .views import view_config

__all__ = [
    "Configurator",
    "GenericConfigurator",
    "configure",
    "exception_handler",
    "resource",
    "resource_view",
    "view_config",
]



"""
The configurator registers routes in a FastAPI application while
adding support for dependency injection during the configuration phase.

FastAPI does not provide any built-in support for dependency injection
during the configuration phase.
Instead, it only resolves dependencies at request time, ensuring they
are dynamically handled per request.

The configurator is designed to handle the setup during the configuration
phase.
"""

import logging
from asyncio import iscoroutine
from collections import defaultdict
from collections.abc import Callable, Sequence
from enum import Enum
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Annotated, Any, Generic, Self, TypeVar

import venusian
from fastapi import Depends, FastAPI, Response
from fastapi import Request as BaseRequest
from fastapi.params import Depends as DependsType
from fastapi.staticfiles import StaticFiles
from fastapi.types import IncEx

from fastlife.adapters.fastapi.request import GenericRequest, Request
from fastlife.adapters.fastapi.routing.route import Route
from fastlife.adapters.fastapi.routing.router import Router
from fastlife.config.openapiextra import OpenApiTag
from fastlife.domain.model.template import InlineTemplate
from fastlife.middlewares.base import AbstractMiddleware
from fastlife.service.check_permission import check_permission
from fastlife.service.csrf import check_csrf
from fastlife.service.registry import DefaultRegistry, TRegistry
from fastlife.settings import Settings
from fastlife.shared_utils.infer import is_inline_template_returned
from fastlife.shared_utils.resolver import (
    resolve,
    resolve_maybe_relative,
)

if TYPE_CHECKING:
    from fastlife.service.locale_negociator import LocaleNegociator  # coverage: ignore
    from fastlife.service.request_factory import (
        RequestFactoryBuilder,  # coverage: ignore
    )
    from fastlife.service.security_policy import AbstractSecurityPolicy
    from fastlife.service.templates import (
        AbstractTemplateRendererFactory,  # coverage: ignore
    )

log = logging.getLogger(__name__)
VENUSIAN_CATEGORY = "fastlife"

venusian_ignored_item = str | Callable[[str], bool]


class ConfigurationError(Exception):
    """
    Error raised during configuration, to avoid errors at runtime.
    """


def rebuild_router(router: Router) -> Router:
    """
    Fix the router.

    FastAPI routers has dependencies that are injected to routes where they are added.

    It means that if you add a dependencies in the router after the route has
    been added, then the dependencies is missing in the route added before.

    To prenvents issues, we rebuild the router route with the dependency.

    :param router: the router to rebuild
    :return: a new router with fixed routes.
    """
    if not router.dependencies:
        return router
    _router = Router(prefix=router.prefix)
    _router.dependencies = router.dependencies
    route: Route
    for route in router.routes:  # type: ignore
        dependencies = [
            dep for dep in route.dependencies if dep not in _router.dependencies
        ]
        _router.add_api_route(
            path=route.path,
            endpoint=route.endpoint,
            response_model=route.response_model,
            status_code=route.status_code,
            tags=route.tags,
            dependencies=dependencies,
            summary=route.summary,
            description=route.description,
            response_description=route.response_description,
            deprecated=route.deprecated,
            methods=route.methods,
            operation_id=route.operation_id,
            response_model_include=route.response_model_include,
            response_model_exclude=route.response_model_exclude,
            response_model_by_alias=route.response_model_by_alias,
            response_model_exclude_unset=route.response_model_exclude_unset,
            response_model_exclude_defaults=route.response_model_exclude_defaults,
            response_model_exclude_none=route.response_model_exclude_none,
            include_in_schema=route.include_in_schema,
            name=route.name,
            openapi_extra=route.openapi_extra,
        )
    return _router


class GenericConfigurator(Generic[TRegistry]):
    """
    Configure and build an application.

    Initialize the app from the settings.
    """

    registry: TRegistry

    def __init__(self, settings: Settings) -> None:
        """
        :param settings: application settings.
        """
        registry_cls = resolve(settings.registry_class)
        self.registry = registry_cls(settings)
        Route._registry = self.registry  # type: ignore

        self.middlewares: list[tuple[type[AbstractMiddleware], Any]] = []
        self.exception_handlers: list[tuple[int | type[Exception], Any]] = []
        self.mounts: list[tuple[str, Path, str]] = []
        self.tags: dict[str, OpenApiTag] = {}

        self.api_title = "OpenAPI"
        self.api_version = "v1"
        self.api_description: str = ""
        self.api_summary: str | None = None
        self.api_swagger_ui_url: str | None = None
        self.api_redoc_url: str | None = None

        self._route_prefix: str = ""
        self._routers: dict[str, Router] = defaultdict(Router)
        self._security_policies: dict[
            str, type[AbstractSecurityPolicy[Any, Any, TRegistry]]
        ] = {}

        self._registered_permissions: set[str] = set()

        self._renderer_globals: dict[str, Any] = {}
        self.scanner = venusian.Scanner(fastlife=self)
        self.include("fastlife.views")
        self.include("fastlife.middlewares")

    @property
    def _current_router(self) -> Router:
        return self._routers[self._route_prefix]

    @property
    def all_registered_permissions(self) -> Sequence[str]:
        """
        Get the list of registered permissions.

        This is usefull for introspection or testing purpose.
        """
        return sorted(self._registered_permissions)

    def build_asgi_app(self) -> FastAPI:
        """
        Build the app after configuration in order to start after beeing configured.

        :return: FastAPI application.
        """

        # register our main template renderer at then end, to ensure that
        # if settings have been manipulated, everything is taken into account.
        for optional_adapter in ("jinjax", "xcomponent"):
            try:
                self.include(
                    f"fastlife.adapters.{optional_adapter}",
                    ignore="fastlife.adapters.jinjax.jinjax_ext.jinjax_doc",
                )
            except ImportError:
                pass

        app = FastAPI(
            title=self.api_title,
            version=self.api_version,
            description=self.api_description,
            summary=self.api_summary,
            dependencies=[Depends(check_csrf())],
            docs_url=self.api_swagger_ui_url,
            redoc_url=self.api_redoc_url,
            lifespan=self.registry.lifespan,
            openapi_tags=[tag.model_dump(by_alias=True) for tag in self.tags.values()]
            if self.tags
            else None,
        )
        app.router.route_class = Route

        for middleware_class, options in self.middlewares:
            app.add_middleware(middleware_class, **options)  # type: ignore

        for status_code_or_exc, exception_handler in self.exception_handlers:
            app.add_exception_handler(status_code_or_exc, exception_handler)

        for prefix, router in self._routers.items():
            app.include_router(rebuild_router(router), prefix=prefix)

        for route_path, directory, name in self.mounts:
            app.mount(route_path, StaticFiles(directory=directory), name=name)
        return app

    def include(
        self,
        module: str | ModuleType,
        route_prefix: str = "",
        ignore: venusian_ignored_item | Sequence[venusian_ignored_item] | None = None,
    ) -> Self:
        """
        Include a module in order to load its configuration.

        It will scan and load all the submodule until you add an ignore rule.

        The `ignore` argument allows you to ignore certain modules.
        If it is a scrint, it can be an absolute module name or a relative
        one, if starts with a dot.

        Here is an example.

        ```python
        from fastlife import Configurator, configure

        def home() -> dict[str, str]:
            return {"hello": "world"}

        @configure
        def includeme(config: Configurator) -> None:
            config.add_route("home", "/", home)
        ```

        :param module: a module to include.
        :param route_prefix: prepend all included route with a prefix
        :param ignore: ignore submodules
        """
        if isinstance(module, str):
            try:
                module = resolve_maybe_relative(module, stack_depth=2)
            except ModuleNotFoundError as exc:
                raise ConfigurationError(f"Can't resolve {module}") from exc

        old, self._route_prefix = self._route_prefix, route_prefix
        try:
            self.scanner.scan(  # type: ignore
                module,
                categories=[VENUSIAN_CATEGORY],
                ignore=ignore,
            )
        finally:
            self._route_prefix = old
        return self

    def set_request_factory(
        self, request_factory: "RequestFactoryBuilder[TRegistry]"
    ) -> Self:
        """Install a request factory, to use a custom request classes."""
        self.registry.request_factory = request_factory(self.registry)
        return self

    def set_locale_negociator(self, locale_negociator: "LocaleNegociator") -> Self:
        """Install a locale negociator for the app."""
        self.registry.locale_negociator = locale_negociator
        return self

    def add_translation_dirs(self, locales_dir: str) -> Self:
        """
        Add a translation directory for localization.

        Usually, the locales is a directory from a package and has to be
        passed with a `:` separator: {package_name}:locales.

        :param locales_dir: the directory contains local inside a python package.
        """
        self.registry.localizer.load(locales_dir)
        return self

    def set_api_documentation_info(
        self,
        title: str,
        version: str,
        description: str,
        *,
        summary: str | None = None,
        swagger_ui_url: str | None = None,
        redoc_url: str | None = None,
    ) -> Self:
        """
        Set your api documentation title for application that expose an API.

        :param title: OpenAPI documentation title
        :param version: OpenAPI api version
        :param description: OpenAPI documentation description. Use markdown here.
        :param summary: OpenAPI documentation summary.
            A short description: text only.
        :param swagger_ui_url: Endpoint for {term}`Swagger UI` served by FastAPI
        :param redoc_url: Endpoint for {term}`Redoc` served by FastAPI
        """
        self.api_title = title
        self.api_version = version
        self.api_description = description
        self.api_summary = summary
        self.api_swagger_ui_url = swagger_ui_url
        self.api_redoc_url = redoc_url
        return self

    def add_openapi_tag(self, tag: OpenApiTag) -> Self:
        """Register a tag description in the {term}`OpenAPI` documentation."""
        if tag.name in self.tags:
            raise ConfigurationError(f"Tag {tag.name} can't be registered twice.")
        self.tags[tag.name] = tag
        return self

    def add_middleware(
        self, middleware_class: type[AbstractMiddleware], **options: Any
    ) -> Self:
        """
        Add a starlette middleware to the FastAPI app.
        """
        self.middlewares.append((middleware_class, options))
        return self

    def set_security_policy(
        self, security_policy: "type[AbstractSecurityPolicy[TRegistry, Any, Any]]"
    ) -> Self:
        """
        Set a security policy for the application.

        ```{important}
        The security policy is **per route_prefix**.
        It means that if the application is splitted via multiple
        route_prefix using the {meth}`Configurator.include`, they
        all have a distinct security policy. A secutity policy has
        to be install by all of those include call.

        :param security_policy: The security policy that will applied for the app
            portion behind the route prefix.
        ```
        """
        self._security_policies[self._route_prefix] = security_policy
        self._current_router.dependencies.append(Depends(security_policy))
        return self

    def add_api_route(
        self,
        name: str,
        path: str,
        endpoint: Callable[..., Any],
        *,
        permission: str | None = None,
        status_code: int | None = None,
        tags: list[str | Enum] | None = None,
        summary: str | None = None,
        description: str | None = None,
        response_description: str = "Successful Response",
        # responses: Dict[Union[int, str], Dict[str, Any]] | None = None,
        deprecated: bool | None = None,
        methods: list[str] | None = None,
        operation_id: str | None = None,
        response_model: Any = None,
        response_model_include: IncEx | None = None,
        response_model_exclude: IncEx | None = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        # response_class: Union[Type[Response], DefaultPlaceholder] = Default(
        #     HTMLResponse
        # ),
        openapi_extra: dict[str, Any] | None = None,
        # generate_unique_id_function: Callable[[APIRoute], str] = Default(
        #     generate_unique_id
        # ),
    ) -> Self:
        """
        Add an API route to the app.

        Fastlife does not use a decorator to attach routes, instead the decorator
        {func}`fastlife.config.configurator.configure` has to be used to
        inject routes inside a method and call the add_route method.

        This route has to be used to add API Route, by API, to expose it in the
        documentation.

        To add a route that serve HTML user the method {meth}`Configurator.add_route`

        :param name: name of the route, used to build route from the helper
            {meth}`fastlife.request.request.Request.url_for` in order to create links.
        :param path: path of the route, use `{curly_brace}` to inject FastAPI Path
            parameters.
        :param endpoint: the function that will reveive the request.
        :param permission: a permission to validate by the security policy.
        :param methods: restrict route to a list of http methods.
        :param description:{term}`OpenAPI` description for the route.
        :param summary: OpenAPI summary for the route.
        :param response_description: OpenAPI description for the response.
        :param operation_id: OpenAPI optional unique string used to identify an
            operation.
        :param tags: OpenAPI tags for the route.
        :param deprecated: OpenAPI deprecated annotation for the route.

        :param response_model_include: customize fields list to include in repsonse.
        :param response_model_exclude: customize fields list to exclude in repsonse.
        :param response_model_by_alias: serialize fields by alias or by name if False.
        :param response_model_exclude_unset: exclude fields that are not explicitly
            set in response.
        :param response_model_exclude_defaults: exclude default value of response
            fields.
        :param response_model_exclude_none: exclude fields instead of serialize to
            null value.
        :param include_in_schema: expose or not the route in the doc.
        :param openapi_extra: OpenAPI documentation extra fields.

        :return: the configurator.
        """
        dependencies: list[DependsType] = []
        if permission:
            self._registered_permissions.add(permission)
            dependencies.append(Depends(check_permission(permission)))

        self._current_router.add_api_route(
            path,
            endpoint,
            response_model=response_model,
            status_code=status_code,
            tags=tags,
            dependencies=dependencies,
            summary=summary,
            description=description,
            response_description=response_description,
            # responses=responses,
            deprecated=deprecated,
            methods=methods,
            operation_id=operation_id or name,
            response_model_include=response_model_include,
            response_model_exclude=response_model_exclude,
            response_model_by_alias=response_model_by_alias,
            response_model_exclude_unset=response_model_exclude_unset,
            response_model_exclude_defaults=response_model_exclude_defaults,
            response_model_exclude_none=response_model_exclude_none,
            include_in_schema=include_in_schema,
            # response_class=response_class,
            name=name,
            openapi_extra=openapi_extra,
            # generate_unique_id_function=generate_unique_id_function,
        )
        return self

    def add_renderer_global(
        self, name: str, value: Any, *, evaluate: bool = True
    ) -> None:
        """
        Add a rendering global value.

        :param name: the name or key of the global value available in the template.
        :param value: a value, or a callable or
            an async function with a request in parameter that will evaluate the value.
        :param evaluate: set to false if you want to inject helper methods in the
            template.
        """
        self._renderer_globals[name] = value, evaluate

    async def _build_renderer_globals(self, request: Request) -> dict[str, Any]:
        """
        Build globals variables accessible in any templates.

        * `request` is the {class}`current request <fastlife.request.request.Request>`
        * `authenticated_user` is used to access to the authenticated user if the
            security policy has been installed.
        * `csrf_token` is used to build for {jinjax:component}`CsrfToken`.
        * `gettext`, `ngettext`, `dgettext`, `dngettext`, `pgettext`, `dpgettext`,
        `npgettext`, `dnpgettext` methods are installed for i18n purpose.
        """
        lczr = request.registry.localizer(request.locale_name)
        custom_globals = {}
        for key, (val, evaluate) in self._renderer_globals.items():
            if evaluate and callable(val):
                val = val(request)
                if iscoroutine(val):
                    val = await val
            custom_globals[key] = val
        return {
            "request": request,
            **lczr.as_dict(),
            **custom_globals,
            **request.renderer_globals,
        }

    def add_route(
        self,
        name: str,
        path: str,
        endpoint: Callable[..., Any],
        *,
        permission: str | None = None,
        template: str | None = None,
        status_code: int | None = None,
        methods: list[str] | None = None,
    ) -> Self:
        """
        Add a route to the app.

        Fastlife does not use the FastAPI decorator to attach routes, instead the
        decorator {func}`@configure <fastlife.config.configurator.configure>` has to
        be used to inject routes inside a method and call the add_route method.
        Or the decorator {func}`@view_config <fastlife.config.views.view_config>`
        can decorate view functions.

        :param name: name of the route, used to build route from the helper
            {meth}`fastlife.request.request.Request.url_for` in order to create links.
        :param path: path of the route, use `{curly_brace}` to inject FastAPI Path
            parameters.
        :param endpoint: the function that will reveive the request.
        :param permission: a permission to validate by the
            {class}`Security Policy <fastlife.service.security_policy.AbstractSecurityPolicy>`.
        :param status_code: customize response status code.
        :param methods: restrict route to a list of http methods.
        :return: the configurator.
        """
        dependencies: list[DependsType] = []
        if permission:
            self._registered_permissions.add(permission)
            dependencies.append(Depends(check_permission(permission)))

        if is_inline_template_returned(endpoint):

            async def render(
                request: Request,
                resp: Annotated["Response | InlineTemplate", Depends(endpoint)],
            ) -> Response:
                if isinstance(resp, Response):
                    return resp

                template = resp.renderer
                globs = await self._build_renderer_globals(request)
                return request.registry.get_renderer(template)(request).render(
                    template,
                    params=resp,
                    globals=globs,
                )

            endpoint = render

        self._current_router.add_api_route(
            path,
            endpoint,
            status_code=status_code,
            dependencies=dependencies,
            methods=methods,
            include_in_schema=False,
            name=name,
        )
        return self

    def add_static_route(
        self, route_path: str, directory: Path, name: str = "static"
    ) -> Self:
        """
        Mount a directory to an http endpoint.

        :param route_path: the root path for the statics.
        :param directory: the directory on the filesystem where the statics files are.
        :param name: a name for the route in the starlette app.
        :return: the configurator

        """
        self.mounts.append((route_path, directory, name))
        return self

    def add_exception_handler(
        self,
        status_code_or_exc: int | type[Exception],
        handler: Callable[..., "Response | InlineTemplate"],
        *,
        status_code: int = 500,
    ) -> Self:
        """
        Add an exception handler the application.

        """

        async def exception_handler(request: BaseRequest, exc: Exception) -> Any:
            # FastAPI exception handler does not provide our request object
            # it seems like it is rebuild from the asgi scope. Even the router
            # class is wrong.
            # Until we store a security policy per rooter, we rebuild an
            # incomplete request here.
            req = GenericRequest[DefaultRegistry, Any, Any](self.registry, request)
            resp = handler(req, exc)
            if isinstance(resp, Response):
                return resp

            return req.registry.get_renderer(resp.renderer)(req).render(
                resp.template,
                params=resp,
                status_code=status_code,
                globals=(await self._build_renderer_globals(req)),
            )

        self.exception_handlers.append((status_code_or_exc, exception_handler))
        return self

    def add_renderer(
        self, file_ext: str, renderer: "AbstractTemplateRendererFactory"
    ) -> Self:
        """
        Add a render for a given file extension.

        :param file_ext: the file extention of your templates.
        :param renderer: the renderer that will render the template.
        """
        # we don't want to expose the renderer publicly as mutable
        self.registry.renderers[f".{file_ext.lstrip('.')}"] = renderer  # type: ignore
        return self

    def add_template_search_path(self, path: str | Path) -> Self:
        """
        Add a template search path directly from the code.

        :param path: template path.
        """
        self.registry.settings.template_search_path = (
            f"{self.registry.settings.template_search_path},{path}"
        )
        return self


class Configurator(GenericConfigurator[DefaultRegistry]):
    """
    Configure and build an application.

    Initialize the app from the settings.
    """


TConfigurator = TypeVar("TConfigurator", bound=GenericConfigurator[Any])


def configure(
    wrapped: Callable[[TConfigurator], None],
) -> Callable[[Any], None]:
    """
    Decorator used to attach route in a submodule while using the configurator.include.
    """

    def callback(
        scanner: venusian.Scanner, name: str, ob: Callable[[venusian.Scanner], None]
    ) -> None:
        if hasattr(scanner, "fastlife"):
            ob(scanner.fastlife)  # type: ignore

    venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
    return wrapped



from pydantic import Field

from .base import Widget


class MFACodeWidget(Widget[str]):
    """
    Widget for MFA code such as TOTP token.
    """

    template = """
    <pydantic_form.Widget :widget_id="id" :removable="removable">
      <div class="pt-4">
        <Label :for="id">{{ gettext(title) }}</Label>
        <pydantic_form.Error :text="error" />
        <Input :name="name" type="text" :id="id" inputmode="numeric"
            autocomplete="one-time-code" :autofocus="autofocus"
          :aria-label="aria_label" :placeholder="placeholder" />
        <pydantic_form.Hint :text="hint" />
      </div>
    </pydantic_form.Widget>
    """

    placeholder: str | None = Field(default=None)
    autofocus: bool = Field(default=True)



"""Handle Union type."""

from collections.abc import Mapping
from types import NoneType
from typing import Any

from pydantic import ValidationError
from pydantic.fields import FieldInfo

from fastlife.adapters.jinjax.widget_factory.base import BaseWidgetBuilder
from fastlife.adapters.jinjax.widgets.base import Widget
from fastlife.adapters.jinjax.widgets.union import UnionWidget
from fastlife.shared_utils.infer import is_complex_type, is_union


class UnionBuilder(BaseWidgetBuilder[Any]):
    """Builder for Union."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for unions Union[A,B], A | B or event Optional[A], A | None"""
        return is_union(typ)

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],
        field: FieldInfo | None,
        value: Any | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> Widget[Any]:
        """Build the widget."""
        types: list[type[Any]] = []
        # required = True
        for typ in field_type.__args__:  # type: ignore
            if typ is NoneType:
                # required = False
                continue
            types.append(typ)  # type: ignore

        if (
            not removable
            and len(types) == 1
            # if the optional type is a complex type,
            and not is_complex_type(types[0])
        ):
            return self.factory.build(  # coverage: ignore
                types[0],
                name=field_name,
                field=field,
                value=value,
                form_errors=form_errors,
                removable=False,
            )
        child = None
        if value:
            for typ in types:
                try:
                    typ(**value)
                except ValidationError:
                    pass
                else:
                    child = self.factory.build(
                        typ,
                        name=field_name,
                        field=field,
                        value=value,
                        form_errors=form_errors,
                        removable=False,
                    )

        widget = UnionWidget[Any](
            name=field_name,
            # we assume those types are BaseModel
            value=child,
            children_types=types,  # type: ignore
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            removable=removable,
            error=form_errors.get(field_name),
        )

        return widget



"""Navigation component."""

from collections.abc import Mapping

from xcomponent import XNode

from fastlife.adapters.xcomponent.catalog import catalog


@catalog.component
def A(
    href: str,
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
    hx_target: str = "#maincontent",
    hx_select: str | None = None,
    hx_swap: str = "innerHTML show:body:top",
    hx_push_url: bool = True,
    hx_get: str | None = None,
    hx_disable: bool | None = None,
    hx_disabled_elt: str | None = None,
) -> str:
    """
    Create html ``<a>`` node with htmx support by default.
    The `hx-get` parameter is set with the href directly unless the
    `disabled-htmx` attribute has been set.

    :param href: Target link.
    :param id: Unique identifier of the element.
    :param class_: CSS class for the node, defaults to
        :attr:`fastlife.template_globals.Globals.A_CLASS`.
    :param hx_target: Target the element for swapping than the one issuing
        the AJAX request.
    :param hx_select: Select the content swapped from the response of the AJAX request.
    :param hx_swap: Specify how the response will be swapped in relative to the target
        of an AJAX request.
    :param hx_push_url: Replace the browser URL with the link.
    :param hx_get: Override the target link only for htmx request for component
        rendering. href will be used if None.
    :param disable_htmx: Do not add any `hx-*` attribute to the link.
    """

    return """
        <a
            href={href}
            id={id}
            hx-disable={hx_disable}
            hx-disabled-elt={hx_disabled_elt}
            hx-get={hx_get or href}
            hx-target={hx_target}
            hx-swap={hx_swap}
            hx-push-url={hx_push_url}
            hx-select={hx_select}
            class={class_ or globals.A_CLASS}
        >
            {children}
        </a>
    """



from pydantic import Field

from .base import Widget


class MFACodeWidget(Widget[str]):
    """
    Widget for MFA code such as TOTP token.
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div class="pt-4">
        <Label for={id}>{ globals.gettext(title) }</Label>
        <OptionalErrorText text={error} />
        <Input name={name} id={id} inputmode="numeric"
            autocomplete="one-time-code" autofocus={autofocus}
            aria-label={aria_label} placeholder={placeholder} />
        <Hint text={hint} />
      </div>
    </Widget>
    """

    placeholder: str | None = Field(default=None)
    autofocus: bool = Field(default=True)



from typing import Annotated

from fastapi import Depends

from fastlife.adapters.fastapi.request import Request
from fastlife.service.translations import Localizer as RequestLocalizer


def get_localizer(request: Request) -> RequestLocalizer:
    """Return the localizer for the given request."""
    return request.registry.localizer(request.locale_name)


Localizer = Annotated[RequestLocalizer, Depends(get_localizer)]
"""Define a localizer"""



"""
Views for pydantic form.

Pydantic form generate form that may contains fields that requires some ajax query.
"""

from typing import Literal

from fastapi import Query
from pydantic.fields import FieldInfo

from fastlife import Configurator, Request, Response, configure
from fastlife.shared_utils.resolver import resolve_extended


async def show_widget(
    typ: str,
    request: Request,
    title: str | None = Query(None),
    name: str | None = Query(None),
    token: str | None = Query(None),
    format: Literal["jinjax", "xcomponent"] | str = Query("jinjax"),
    removable: bool = Query(False),
) -> Response:
    """
    This views is used by pydantic_form to generate a nested field asynchronously.
    """
    model_cls = resolve_extended(typ)
    field = None
    if title:
        field = FieldInfo(title=title)

    if format == "jinjax":
        # XXX
        format = request.registry.settings.jinjax_file_ext
    rndr = request.registry.get_renderer(f".{format}")(request)
    lczr = request.registry.localizer(request.locale_name)
    rndr.globals.update({"request": request, **lczr.as_dict()})
    data = rndr.pydantic_form_field(
        model=model_cls,  # type: ignore
        name=name,
        token=token,
        removable=removable,
        field=field,
    )
    return Response(data, headers={"Content-Type": "text/html"})


@configure
def includeme(config: Configurator) -> None:
    route_prefix = config.registry.settings.fastlife_route_prefix
    config.add_route(
        "fl-pydantic-form-widget",
        f"{route_prefix}/pydantic-form/widgets/{{typ}}",
        show_widget,
        methods=["GET"],
    )



from importlib import metadata

__version__ = metadata.version("fastlifeweb")


from .adapters.fastapi.form import form_model
from .adapters.fastapi.localizer import Localizer
from .adapters.fastapi.request import (
    AnyRequest,
    Registry,
    Request,
    get_registry,
    get_request,
)
from .config import (
    Configurator,
    GenericConfigurator,
    configure,
    exception_handler,
    resource,
    resource_view,
    view_config,
)
from .domain.model.asgi import ASGIRequest, ASGIResponse
from .domain.model.form import FormModel
from .domain.model.request import GenericRequest
from .domain.model.response import RedirectResponse, Response
from .domain.model.security_policy import (
    Allowed,
    Anonymous,
    Authenticated,
    AuthenticationState,
    Denied,
    Forbidden,
    HasPermission,
    NoMFAAuthenticationState,
    PendingMFA,
    PreAuthenticated,
    TClaimedIdentity,
    TIdentity,
    Unauthenticated,
    Unauthorized,
)
from .domain.model.template import JinjaXTemplate, XTemplate

# from .request.form_data import model
from .service.registry import DefaultRegistry, GenericRegistry, TRegistry, TSettings
from .service.request_factory import RequestFactory
from .service.security_policy import (
    AbstractNoMFASecurityPolicy,
    AbstractSecurityPolicy,
    InsecurePolicy,
)
from .service.translations import TranslatableStringFactory
from .settings import Settings

__all__ = [
    # Config
    "GenericConfigurator",
    "GenericRegistry",
    "Registry",
    "Settings",
    "configure",
    "view_config",
    "exception_handler",
    "resource",
    "resource_view",
    "Configurator",
    "DefaultRegistry",
    "TSettings",
    "TRegistry",
    "get_registry",
    # Form
    "FormModel",
    "form_model",
    # Request
    "GenericRequest",
    "AnyRequest",
    "Request",
    "get_request",
    # Request Factory
    "ASGIRequest",
    "ASGIResponse",
    "RequestFactory",
    # Response
    "Response",
    "RedirectResponse",
    # Security
    "AbstractSecurityPolicy",
    "AbstractNoMFASecurityPolicy",
    "HasPermission",
    "Unauthenticated",
    "PreAuthenticated",
    "Allowed",
    "Denied",
    "Unauthorized",
    "Forbidden",
    "InsecurePolicy",
    "Anonymous",
    "PendingMFA",
    "Authenticated",
    "AuthenticationState",
    "NoMFAAuthenticationState",
    "TClaimedIdentity",
    "TIdentity",
    # Template
    "JinjaXTemplate",
    "XTemplate",
    # i18n
    "Localizer",
    "TranslatableStringFactory",
]



"""Abstract class for builder."""

import abc
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from pydantic.fields import FieldInfo

from fastlife.adapters.jinjax.widgets.base import Widget

if TYPE_CHECKING:
    from fastlife.adapters.jinjax.widget_factory.factory import (  # coverage: ignore
        WidgetFactory,
    )

T = TypeVar("T")


class BaseWidgetBuilder(abc.ABC, Generic[T]):
    """Base class for the builder of widget."""

    def __init__(self, factory: "WidgetFactory") -> None:
        self.factory = factory

    @abc.abstractmethod
    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """Return true if the builder accept to build a widget for this type."""

    @abc.abstractmethod
    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],
        field: FieldInfo | None,
        value: T | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> Widget[T]:
        """Build the widget"""



"""Handle EmailStr pydantic type."""

from collections.abc import Mapping
from typing import Any

from pydantic.fields import FieldInfo
from pydantic.networks import EmailStr

from fastlife.adapters.jinjax.widget_factory.base import BaseWidgetBuilder
from fastlife.adapters.jinjax.widgets.text import TextWidget
from fastlife.domain.model.types import Builtins


class EmailStrBuilder(BaseWidgetBuilder[Builtins]):
    """Builder for Pydantic EmailStr."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for EmailStr."""
        return issubclass(typ, EmailStr)  # type: ignore

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],
        field: FieldInfo | None,
        value: Builtins | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> TextWidget:
        """Build the widget."""
        return TextWidget(
            name=field_name,
            input_type="email",
            placeholder=str(field.examples[0]) if field and field.examples else None,
            removable=removable,
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            token=self.factory.token,
            value=str(value),
            error=form_errors.get(field_name),
            autocomplete="email",
        )



"""
Initialize a session.


The session :attr:`fastlife.config.settings.Settings.session_secret_key` must
be set in order to create a session.

This secret is used to sign session content in order to prevent malicious user
to write their own session content. Note that the provided session implementation
does not cipher session content, it just sign. No secret should be placed in the
session.
"""

from fastlife import Configurator, configure
from fastlife.shared_utils.resolver import resolve

from .middleware import SessionMiddleware
from .serializer import AbsractSessionSerializer

__all__ = ["AbsractSessionSerializer", "SessionMiddleware"]


@configure
def includeme(config: Configurator) -> None:
    settings = config.registry.settings
    session_serializer = resolve(settings.session_serializer)
    if settings.session_secret_key:
        config.add_middleware(
            SessionMiddleware,
            cookie_name=settings.session_cookie_name,
            secret_key=settings.session_secret_key,
            duration=settings.session_duration,
            cookie_path=settings.session_cookie_path,
            cookie_same_site=settings.session_cookie_same_site,
            cookie_secure=settings.session_cookie_secure,
            cookie_domain=settings.session_cookie_domain,
            serializer=session_serializer,
        )



from collections.abc import Sequence
from typing import Any

from pydantic import Field, SecretStr

from fastlife.adapters.xcomponent.catalog import catalog
from fastlife.domain.model.types import Builtins

from .base import Widget


@catalog.function
def is_str(value: Any) -> bool:
    return isinstance(value, str)


class TextWidget(Widget[Builtins]):
    """
    Widget for text like field (email, ...).
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div class="pt-4">
        <Label for={id}>{globals.gettext(title)}</Label>
        <OptionalErrorText text={error} />
        <Input name={name} value={value} type={input_type} id={id}
          aria-label={aria_label} placeholder={placeholder}
         autocomplete={autocomplete} />
        <Hint text={hint} />
      </div>
    </Widget>
    """

    input_type: str = Field(default="text")
    """type attribute for the Input component."""
    placeholder: str | None = Field(default=None)
    """placeholder attribute for the Input component."""
    autocomplete: str | None = Field(default=None)
    """autocomplete attribute for the Input component."""


class PasswordWidget(Widget[SecretStr]):
    """
    Widget for password fields.
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div class="pt-4">
        <Label for={id}>{globals.gettext(title)}</Label>
        <OptionalErrorText text={error} />
        <Password name={name} type={input_type} id={id}
          autocomplete={
            if new_password {
              <>new-password</>
            }
            else {
              <>current-password</>
            }
          }
          aria-label={aria_label} placeholder={placeholder} />
        <Hint text={hint} />
      </div>
    </Widget>
    """

    input_type: str = Field(default="password")
    """type attribute for the Input component."""
    placeholder: str | None = Field(default=None)
    """placeholder attribute for the Input component."""
    new_password: bool = Field(default=False)
    """
    Adapt autocomplete behavior for browsers to hint existing or generate password.
    """


class TextareaWidget(Widget[str | Sequence[str]]):
    """
    Render a Textearea for a string or event a sequence of string.

    ```
    from fastlife.adapters.jinjax.widgets.base import CustomWidget
    from fastlife.adapters.jinjax.widgets.text import TextareaWidget
    from pydantic import BaseModel, Field, field_validator

    class TaggedParagraphForm(BaseModel):
        paragraph: Annotated[str, CustomWidget(TextareaWidget)] = Field(...)
        tags: Annotated[Sequence[str], CustomWidget(TextareaWidget)] = Field(
            default_factory=list,
            title="Tags",
            description="One tag per line",
        )

        @field_validator("tags", mode="before")
        def split(cls, s: Any) -> Sequence[str]:
            return s.split() if s else []
    ```
    """

    template = """
    <Widget widget_id={id} removable={removable}>
      <div class="pt-4">
        <Label for={id}>{globals.gettext(title)}</Label>
        <OptionalErrorText text={error} />
        <Textarea name={name} id={id} aria-label={aria_label}>
            {
              if is_str(value) {
                value
              }
              else {
                for v in value {
                  v + "\n"
                }
              }
            }
        </Textarea>
        <Hint text={hint} />
      </div>
    </Widget>
    """



"""
Template rending based on JinjaX.
"""

import ast
import logging
import re
from pathlib import Path
from typing import cast

from jinja2 import Template
from jinjax.component import RX_ARGS_START, RX_META_HEADER, Component

from .docstring import generate_docstring

log = logging.getLogger(__name__)

RX_DOC_START = re.compile(r"{#-?\s*doc\s+")
RX_CONTENT = re.compile(r"\{\{-?\s*content\s*-?\}\}", re.DOTALL)
RX_COMMENT_REPLACE = re.compile(r"{#[^#]+#}")


SIGNATURE = """
def component({args}):
    '''
    {docstring}
    '''
    ...
"""


def has_content(source: str) -> bool:
    nocomment = RX_COMMENT_REPLACE.sub("", source)
    return len(RX_CONTENT.findall(nocomment)) > 0


class InspectableComponent(Component):
    """
    JinjaX Components override that support intropspection.

    The Component class of JinjaX does not support documentation by default.
    """

    __slots__ = (
        "css",
        "js",
        "mtime",
        "name",
        "optional",
        "path",
        "prefix",
        "required",
        "source",
        "tmpl",
        "url_prefix",
    )

    def __init__(
        self,
        *,
        name: str,
        prefix: str = "",
        url_prefix: str = "",
        source: str = "",
        mtime: float = 0,
        tmpl: "Template | None" = None,
        path: "Path | None" = None,
    ) -> None:
        super().__init__(
            name=name,
            prefix=prefix,
            url_prefix=url_prefix,
            source=source,
            mtime=mtime,
            tmpl=tmpl,
            path=path,
        )
        self.source = source

    def as_def(self) -> ast.FunctionDef:
        """Return the component definition as a python function definition."""
        signature = "def component(): pass"
        match = RX_META_HEADER.match(self.source)
        if match:
            headers = match.group(0)
            header = headers.split("#}")[:-1]
            docstring = ""

            expr = None
            while header:
                item = header.pop(0).strip(" -\n")

                expr = self.read_metadata_item(item, RX_ARGS_START)  # type: ignore
                doc = self.read_metadata_item(item, RX_DOC_START)  # type: ignore
                if doc:
                    docstring += f"    {doc.strip()}\n"
                    continue

            sigargs = f"*, {expr}" if expr else ""
            signature = SIGNATURE.format(args=sigargs, docstring=docstring or "")

        astree = ast.parse(signature)
        return cast(ast.FunctionDef, astree.body[0])

    def build_docstring(self) -> str:
        """Build a rst docstring for the jinjax component."""
        func_def = self.as_def()
        prefix = f"{self.prefix}." if self.prefix else ""
        ret = ".. jinjax:component:: " + generate_docstring(
            func_def, f"{prefix}{self.name}", has_content(self.source)
        )
        return ret



"""Handle Sequence type."""

from collections.abc import Mapping, MutableSequence, Sequence
from typing import Any

from pydantic.fields import FieldInfo

from fastlife.adapters.jinjax.widget_factory.base import BaseWidgetBuilder
from fastlife.adapters.jinjax.widgets.base import Widget
from fastlife.adapters.jinjax.widgets.sequence import SequenceWidget


class SequenceBuilder(BaseWidgetBuilder[Sequence[Any]]):
    """Builder for Sequence values."""

    def accept(self, typ: type[Any], origin: type[Any] | None) -> bool:
        """True for Sequence, MutableSequence or list"""
        return origin is Sequence or origin is MutableSequence or origin is list

    def build(
        self,
        *,
        field_name: str,
        field_type: type[Any],
        field: FieldInfo | None,
        value: Sequence[Any] | None,
        form_errors: Mapping[str, Any],
        removable: bool,
    ) -> Widget[Sequence[Any]]:
        """Build the widget."""
        typ = field_type.__args__[0]  # type: ignore
        value = value or []
        items: Sequence[Any] = [
            self.factory.build(
                typ,  # type: ignore
                name=f"{field_name}.{idx}",
                value=v,
                field=field,
                form_errors=form_errors,
                removable=True,
            )
            for idx, v in enumerate(value)
        ]
        return SequenceWidget[Any](
            name=field_name,
            title=field.title or "" if field else "",
            hint=field.description if field else None,
            aria_label=(
                field.json_schema_extra.get("aria_label")  # type:ignore
                if field and field.json_schema_extra
                else None
            ),
            value=items,
            item_type=typ,  # type: ignore
            token=self.factory.token,
            removable=removable,
            error=form_errors.get(field_name),
        )



"""
Fastlife HTTP middlewares are Starlette HTTP Middleware.

HTTP middleware is a function that processes requests before they reach
the view function or modifies responses after the view has processed
the request.
"""



"""Type inference."""

import inspect
from collections.abc import Callable
from types import UnionType
from typing import Any, Union, get_args, get_origin

from pydantic import BaseModel

from fastlife.domain.model.template import InlineTemplate


def is_complex_type(typ: type[Any]) -> bool:
    """
    Used to detect complex type such as Mapping, Sequence and pydantic BaseModel.

    This method cannot be used outside pydantic serialization.
    """
    return bool(get_origin(typ) or issubclass(typ, BaseModel))


def is_union(typ: type[Any]) -> bool:
    """Used to detect unions like Optional[T], Union[T, U] or T | U."""
    type_origin = get_origin(typ)
    if type_origin:
        if type_origin is Union:  # Optional[T]
            return True

        if type_origin is UnionType:  # T | U
            return True
    return False


def is_inline_template_returned(endpoint: Callable[..., Any]) -> bool:
    """Test if a view, the endpoint return a template."""
    signature = inspect.signature(endpoint)
    return_annotation = signature.return_annotation

    if isinstance(return_annotation, type) and issubclass(
        return_annotation, InlineTemplate
    ):
        return True

    if is_union(return_annotation):
        return any(
            isinstance(arg, type) and issubclass(arg, InlineTemplate)
            for arg in get_args(return_annotation)
        )

    return False



"""Testing your application."""

from collections.abc import Mapping, MutableMapping
from typing import Any, Literal, Self
from urllib.parse import urlencode

import bs4
import httpx
from fastapi.testclient import TestClient
from multidict import MultiDict

from fastlife.domain.model.asgi import ASGIApp
from fastlife.middlewares.session.serializer import AbsractSessionSerializer
from fastlife.settings import Settings
from fastlife.shared_utils.resolver import resolve
from fastlife.testing.dom import Element
from fastlife.testing.form import WebForm
from fastlife.testing.session import Session

CookieTypes = httpx._types.CookieTypes  # type: ignore
Cookies = httpx._models.Cookies  # type: ignore


class WebResponse:
    """Represent an http response made by the WebTestClient browser."""

    def __init__(self, client: "WebTestClient", origin: str, response: httpx.Response):
        self._client = client
        self._response = response
        self._origin = origin
        self._html: bs4.BeautifulSoup | None = None
        self._form: WebForm | None = None

    @property
    def status_code(self) -> int:
        """Http status code."""
        return self._response.status_code

    @property
    def is_redirect(self) -> bool:
        """True for any kind of http redirect status."""
        return 300 <= self._response.status_code < 400

    @property
    def content_type(self) -> str:
        """Get the content type of the response, from the header."""
        return self._response.headers.get("content-type", "").split(";").pop(0)

    @property
    def headers(self) -> httpx.Headers:
        """All http headers of the response."""
        return self._response.headers

    @property
    def text(self) -> str:
        """Http response body."""
        return self._response.text

    @property
    def html(self) -> Element:
        """Http response body as an Element."""
        if self._html is None:
            self._html = bs4.BeautifulSoup(self._response.text, "html.parser")
        return Element(self._client, self._html)

    @property
    def html_body(self) -> Element:
        """The body element of the html response."""
        body = self.html.by_node_name("body")
        assert len(body) == 1, "body element not found or multiple body found"
        return body[0]

    @property
    def form(self) -> WebForm:
        """The form element of the html response."""
        if self._form is None:
            form = self.html.form
            assert form is not None, "form element not found"
            self._form = WebForm(self._client, self._origin, form)
        return self._form

    def by_text(
        self, text: str, *, node_name: str | None = None, position: int | None = None
    ) -> Element | None:
        """Search a dom element by its text."""
        return self.html.by_text(text, node_name=node_name, position=position)

    def by_label_text(self, text: str) -> Element | None:
        """Search a dom element by its associated label text."""
        return self.html.by_label_text(text)

    def by_node_name(
        self, node_name: str, *, attrs: dict[str, str] | None = None
    ) -> list[Element]:
        """List dom element having the given node name, and eventually attributes."""
        return self.html.by_node_name(node_name, attrs=attrs)


class WebTestClient:
    """The fake browser used for testing purpose."""

    def __init__(
        self,
        app: ASGIApp,
        *,
        settings: Settings | None = None,
        cookies: CookieTypes | None = None,
    ) -> None:
        self.app = app
        if settings is None:
            settings = Settings()
            settings.domain_name = settings.domain_name or "testserver.local"
        self.testclient = TestClient(
            app, base_url=f"http://{settings.domain_name}", cookies=cookies or {}
        )
        self.settings = settings
        self.session_serializer: AbsractSessionSerializer = resolve(
            settings.session_serializer
        )(
            settings.session_secret_key,
            int(settings.session_duration.total_seconds()),
        )

    @property
    def cookies(self) -> Cookies:
        """HTTP Cookies."""
        return self.testclient.cookies

    @property
    def session(self) -> MutableMapping[str, Any]:
        """Server session stored in a cookies."""
        return Session(self)

    def request(
        self,
        method: Literal["GET", "POST", "DELETE"],
        url: str,
        *,
        content: str | None = None,
        headers: Mapping[str, str] | None = None,
        max_redirects: int = 0,
    ) -> WebResponse:
        """Perform http requests."""
        rawresp = self.testclient.request(
            method=method,
            url=url,
            headers=headers,
            content=content,
            follow_redirects=False,
        )
        resp = WebResponse(
            self,
            url,
            rawresp,
        )
        if resp.is_redirect and max_redirects > 0:
            if resp.status_code != 307:
                method = "GET"
                headers = None
                content = None
            url = resp.headers.get("location")
            if "HX-Redirect" in resp.headers:
                # Redirection requested to the browser from an AJAX request.
                url = resp.headers["HX-Redirect"]
            assert url, "Redirect response without a redirection"
            return self.request(
                method=method,
                url=url,
                content=content,
                headers=headers,
                max_redirects=max_redirects - 1,
            )
        if "HX-Redirect" in resp.headers and max_redirects > 0:
            return self.request(
                method="GET",
                url=resp.headers["HX-Redirect"],
                content=None,
                headers=headers,
                max_redirects=max_redirects - 1,
            )

        return resp

    def get(self, url: str, follow_redirects: bool = True) -> WebResponse:
        """Perform http GET request."""
        return self.request(
            "GET",
            url,
            max_redirects=int(follow_redirects) * 10,
        )

    def delete(self, url: str, follow_redirects: bool = True) -> WebResponse:
        """Perform http DELETE request."""
        return self.request(
            "DELETE",
            url,
            max_redirects=int(follow_redirects) * 10,
        )

    def post(
        self,
        url: str,
        data: MultiDict[str],
        *,
        headers: Mapping[str, Any] | None = None,
        follow_redirects: bool = True,
    ) -> WebResponse:
        """Perform http POST request in "application/x-www-form-urlencoded" format."""
        if headers is None:
            headers = {}
        return self.request(
            "POST",
            url,
            content=urlencode(data),
            headers={"Content-Type": "application/x-www-form-urlencoded", **headers},
            max_redirects=int(follow_redirects) * 10,
        )

    def __enter__(self) -> Self:
        self.testclient.__enter__()
        return self

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        self.testclient.__exit__()



"""Customize error pages."""

from collections.abc import Callable
from typing import Any

import venusian

from .configurator import VENUSIAN_CATEGORY, Configurator


def exception_handler(
    exception: type[Exception],
    *,
    status_code: int | None = None,
) -> Callable[..., Any]:
    """
    A decorator function to add an exception handler in the app.

    :param methods: restrict route to a list of http methods.

    :return: the configuration callback.
    """

    def configure(
        wrapped: Callable[..., Any],
    ) -> Callable[..., Any]:
        def callback(
            scanner: venusian.Scanner, name: str, ob: Callable[..., Any]
        ) -> None:
            if not hasattr(scanner, VENUSIAN_CATEGORY):
                return  # coverage: ignore
            config: Configurator = getattr(scanner, VENUSIAN_CATEGORY)
            config.add_exception_handler(
                exception,
                wrapped,
                **({} if status_code is None else {"status_code": status_code}),
            )

        venusian.attach(wrapped, callback, category=VENUSIAN_CATEGORY)  # type: ignore
        return wrapped

    return configure



from collections.abc import Mapping
from typing import Literal

from xcomponent import XNode

from fastlife.adapters.xcomponent.catalog import catalog


@catalog.component
def Form(
    children: XNode,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
    method: Literal["get", "post"] | None = None,
    action: str | None = None,
    hx_target: str | None = None,
    hx_select: str | None = None,
    hx_swap: str | None = None,
    hx_post: str | Literal[True] | None = None,
    hx_disable: Literal[True] | None = None,
) -> str:
    return """
        <form
            id={id}
            class={class_ or globals.FORM_CLASS}
            hx-disable={hx_disable}
            hx-post={hx_post}
            hx-select={hx_select}
            hx-swap={hx_swap}
            hx-target={hx_target}
            action={action}
            method={method}
            >
            <CsrfToken />
            { children }
        </form>
    """


@catalog.component
def CsrfToken(globals: dict[str, str]) -> str:
    return """
        <Hidden
            name={globals.request.csrf_token.name}
            value={globals.request.csrf_token.value}
            />
    """


@catalog.component
def Input(
    name: str,
    value: str = "",
    type: str = "text",
    id: str | None = None,
    class_: str | None = None,
    aria_label: str | None = None,
    placeholder: str | None = None,
    inputmode: Literal[
        "none", "text", "decimal", "numeric", "tel", "search", "email", "url"
    ]
    | None = None,
    autocomplete: Literal[
        "on",
        "off",
        "name",
        "username",
        "current-password",
        "new-password",
        "one-time-code",
        "email",
        "tel",
        "organization",
        "street-address",
        "address-line1",
        "address-line2",
        "address-line3",
        "postal-code",
        "country",
        "country-name",
        "cc-name",
        "cc-number",
        "cc-exp",
        "cc-csc",
        "tel-country-code",
        "tel-national",
        "tel-area-code",
        "tel-local",
        "tel-extension",
        "bday",
        "bday-day",
        "bday-month",
        "bday-year",
        "transaction-amount",
        "transaction-currency",
    ]
    | None = None,
    autofocus: bool = False,
) -> str:
    return """
    <input
        type={type}
        name={name}
        value={value}
        id={id}
        aria-label={ aria_label }
        placeholder={ placeholder }
        inputmode={inputmode}
        autocomplete={autocomplete}
        class={class_ or globals.INPUT_CLASS}
        autofocus={autofocus}
    />
    """


@catalog.component
def Hidden(
    name: str,
    value: str,
    id: str | None = None,
) -> str:
    return """
        <input name={name} value={value} type="hidden" id={id} />
    """


@catalog.component
def Button(
    children: XNode,
    globals: Mapping[str, str],
    type: Literal["submit", "button", "reset"] = "submit",
    id: str | None = None,
    class_: str | None = None,
    name: str = "action",
    value: str = "submit",
    hidden: bool = False,
    aria_label: str | None = None,
    onclick: str | None = None,
    hx_target: str | None = None,
    hx_swap: str | None = None,
    hx_select: str | None = None,
    hx_on_after_request: str | None = None,
    hx_vals: str | None = None,
    hx_confirm: str | None = None,
    hx_get: str | None = None,
    hx_post: str | None = None,
    hx_put: str | None = None,
    hx_patch: str | None = None,
    hx_delete: str | None = None,
    hx_params: str | None = None,
    hx_push_url: bool = False,
    full_width: bool = False,
) -> str:
    return """
    <button
        type={type}
        name={name}
        value={value}
        id={id}
        hx-target={hx_target}
        hx-swap={hx_swap}
        hx-select={hx_select}
        onclick={onclick}
        hx-on::after-request={hx_on_after_request}
        hx-vals={hx_vals}
        hx-confirm={hx_confirm}
        hx-get={hx_get}
        hx-post={hx_post}
        hx-put={hx_put}
        hx-patch={hx_patch}
        hx-delete={hx_delete}
        hx-push-url={hx_push_url}
        hx-params={hx_params}
        aria-label={aria_label}
        class={
            if full_width {
                "w-full " + (class_ or globals.BUTTON_CLASS)
            }
            else {
                class_ or globals.BUTTON_CLASS
            }
        }
        hidden={hidden}
        >
        {children}
    </button>
    """


@catalog.component
def Checkbox(
    name: str,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
    value: str | None = None,
    checked: bool = False,
) -> str:
    """
    Create html ``<input type="checkbox" />`` node.

    :param name: Name of the checkbox
    :param id: unique identifier of the element
    :param class_: css class for the node, defaults to
                  :attr:`fastlife.template_globals.Globals.CHECKBOX_CLASS`
    :param value: http submitted value if the checkbox is checked
    :param checked: Initialized the checkbox as ticked
    """
    return """
        <input
            name={name}
            type="checkbox"
            id={id}
            value={value}
            class={class_ or globals.CHECKBOX_CLASS }
            checked={checked}
            />
    """


@catalog.component
def Password(
    name: str,
    globals: Mapping[str, str],
    id: str | None = None,
    class_: str | None = None,
    aria_label: str | None = None,
    placeholder: str | None = None,
    autocomplete: Literal[
        "on",
        "off",
        "current-password",
        "new-password",
    ]
    | None = None,
    inputmode: Literal[
        "none",
        "text",
        "numeric",
    ]
    | None = None,
    minlength: int | None = None,
    maxlength: int | None = None,
    pattern: str | None = None,
    autofocus: bool = False,
    required: bool = False,
    readonly: bool = False,
) -> str:
    """
    Produce ``<input type="password">`` node.

    :param name: submitted name in the form
    :param id: unique identifier of the element
    :param class_: css class for the node, defaults to
                  :attr:`fastlife.template_globals.Globals.INPUT_CLASS`
    :param aria_label: aria-label
    :param placeholder: brief hint to the user as to what kind of information
                        is expected in the field
    :param autocomplete: Define autocomplete mode
    :param inputmode: Define a virtual keyboard layout
    :param minlength: Minimum length
    :param maxlength: Maximum length
    :param pattern: Must match a pattern
    :param autofocus: Give the focus
    :param required: Mark as required field
    :param readonly: Mark as readonly field
    """
    return """
    <input
        name={name}
        type="password"
        id={id}
        aria-label={aria_label}
        placeholder={placeholder}
        autocomplete={autocomplete}
        inputmode={inputmode}
        minlength={minlength}
        maxlength={maxlength}
        pattern={pattern}
        class={class_ or globals.INPUT_CLASS}
        autofocus={autofocus}
        required={required}
        readonly={readonly}
    />
    """


@catalog.component
def Label(
    children: XNode,
    globals: Mapping[str, str],
    for_: str | None = None,
    id: str | None = None,
    class_: str | None = None,
) -> str:
    """
    Produce ``<label>`` node.

    :param for_: unique identifier of the target element.
    :param id: unique identifier of the element.
    :param class_: css class for the node, defaults to
                  :attr:`fastlife.template_globals.Globals.LABEL_CLASS`.
    """
    return """
    <label for={for_} class={class_ or globals.LABEL_CLASS} id={id}>
        {children}
    </label>
    """


@catalog.component
def Option(
    children: XNode,
    globals: Mapping[str, str],
    value: str,
    id: str | None = None,
    selected: bool = False,
) -> str:
    """
    Produce ``<option>`` node.

    :param value: posted value after submitted the selected value
    :param id: unique identifier of the element
    :param selected: Used to select the option while rendering the form
    """
    return """
    <option value={value} id={id} selected={selected}>
        {children}
    </option>
    """


@catalog.component
def Select(
    children: XNode,
    globals: Mapping[str, str],
    name: str,
    id: str | None = None,
    class_: str | None = None,
    multiple: bool = False,
) -> str:
    """
    Create html ``<select>`` node.

    :param name: name of the submitted
    :param id: unique identifier of the element
    :param class_: css class for the node, defaults to
                  :attr:`fastlife.template_globals.Globals.SELECT_CLASS`
    :param multiple: Mark as multiple
    """
    return """
    <select name={name} id={id} class={class_ or globals.SELECT_CLASS}
            multiple={multiple}>
        {children}
    </select>
    """


@catalog.component
def Radio(
    globals: Mapping[str, str],
    label: str,
    name: str,
    value: str,
    id: str | None = None,
    checked: bool = False,
    disabled: bool = False,
    onclick: str | None = None,
    div_class: str | None = None,
    class_: str | None = None,
    label_class: str | None = None,
) -> str:
    """
    Produce a ``<input type="radio">`` with its associated label inside a div.

    :param label: label text associated to the radio
    :param name: name of the submitted
    :param value: value that will be submitted if selected
    :param id: unique identifier of the element
    :param checked: Tick the radio button
    :param disabled: Mark the radio button as disabled
    :param onclick: execute some javascript while clicking
    :param div_class: css class for the div node, defaults to
                     :attr:`fastlife.template_globals.Globals.RADIO_DIV_CLASS`
    :param class_: css class for the input node, defaults to
                   :attr:`fastlife.template_globals.Globals.RADIO_INPUT_CLASS`
    :param label_class: css class for the label node, defaults to
                        :attr:`fastlife.template_globals.Globals.RADIO_LABEL_CLASS`
    """

    return """
    <div class={div_class or globals.RADIO_DIV_CLASS}>
        <input type="radio" name={name} id={id} value={value}
            class={class_ or globals.RADIO_INPUT_CLASS}
            onclick={onclick}
            checked={checked}
            disabled={disabled} />
        <Label for={id} class={label_class or globals.RADIO_LABEL_CLASS}>
            {label}
        </Label>
    </div>
    """


@catalog.component
def Textarea(
    children: XNode,
    globals: Mapping[str, str],
    name: str,
    id: str | None = None,
    class_: str | None = None,
    aria_label: str | None = None,
    placeholder: str | None = None,
) -> str:
    """
    html ``<textarea>`` node.

    :param name: name of the submitted
    :param id: unique identifier of the element
    :param class_: css class for the node, defaults to
                   :attr:`fastlife.template_globals.Globals.INPUT_CLASS`
    :param aria_label: aria-label
    :param placeholder: brief hint to the user as to what kind of information
                        is expected in the field
    """

    return """
    <textarea
        name={name}
        id={id}
        aria-label={aria_label}
        placeholder={placeholder}
        class={class_ or globals.INPUT_CLASS}>
        {children}
    </textarea>
    """



"""HTTP Request representation in a python object."""

from typing import Annotated, Any

from fastapi import Request as FastAPIRequest
from fastapi.params import Depends

from fastlife.domain.model.request import GenericRequest
from fastlife.service.registry import DefaultRegistry


def get_request(request: FastAPIRequest) -> GenericRequest[Any, Any, Any]:
    """Return the Fastlife Request object."""
    return request  # type: ignore


Request = Annotated[GenericRequest[DefaultRegistry, Any, Any], Depends(get_request)]
"""A request that is associated to the default registry."""
# FastAPI handle its Request objects using a lenient_issubclass,
# basically a issubclass(Request), does not work with Generic[T].


AnyRequest = Annotated[GenericRequest[Any, Any, Any], Depends(get_request)]
"""A request version that is associated to the any registry."""


def get_registry(request: Request) -> DefaultRegistry:
    """Return the Fastlife Registry object."""
    return request.registry


Registry = Annotated[DefaultRegistry, Depends(get_registry)]
"""FastAPI dependency to access to the global registry."""
